from munch import Munch

from infi.shell.commands import commandgroup, command, render
from infi.shell.types import Str, YesNo
from infi.shell.utils import strlist
from infi.shell.exceptions import BadValue, BadArguments

from ..types import ConfigCategory, ConfigParameter
from ..options import Grep, InvertedGrep, OutputFormat
from .. import roles, render_collection
from . import ConfigCommandsBase

def _convert_boolean(str_value):
    str_value = str_value.lower().strip()
    if str_value in ("true", "1", "yes"):
        return True
    if str_value in ("false", "0", "no"):
        return False
    raise BadValue()

@commandgroup(parent=ConfigCommandsBase, provides=(ConfigParameter,))
class ConfigCommands(ConfigCommandsBase):

    _TYPES = {key : (value, converter)
              for value, keys, converter in (
                  ("Number",   (int, long, float),          None),
                  ("Text",     (basestring, str, unicode),  None),
                  ("Boolean",  (bool,),                     _convert_boolean),
                  ("List",     (list, tuple, set),          None),
                  )
              for key in keys}
    # Definition of validations of configuration values. When setting
    # configuration values using the 'config.set' command, the validation
    # method which corresponds with the parameter's name is invoked. The method
    # should return either None if the value is valid, or an error string if
    # the value is invalid.
    # It's important to keep in mind that the purpose of this validation code
    # is to work around problems with the management configuration API.
    # Generally, InfiniShell should contain as little application logic as
    # possible, and should not be coupled with management code.
    _VALIDATION_PRESETS = Munch(
            NOT_EMPTY = lambda value: None if value else "Empty value not allowed",
            )
    _VALIDATIONS = {
            "mgmt/environment.dns_servers" : _VALIDATION_PRESETS.NOT_EMPTY,
            }

    @command(
        description="Display system configuration parameters",
        types=(ConfigCategory, ConfigParameter,),
        substitute=dict(category_name="category"),
        options=(Grep(), InvertedGrep(), OutputFormat()),
        )
    @render_collection(
        entity=None,
        fields=strlist("parameter value"),
        detailed_fields=strlist("parameter value"),
        detailed_individually=True,
        )
    @roles.INFINIDAT
    def query(self, _api, category_name=None, parameter=None):
        sort_key = lambda p: p.parameter
        if category_name and parameter:
            raise BadArguments("Category and parameter cannot be specified together")
        if category_name and not parameter:
            category = _api.get("{}/{}".format(self._BASE_URI, category_name))
            parameters = sorted(self._flatten_category({ category_name : category }), key=sort_key)
        elif not category_name and parameter:
            value = _api.get("{}/{}".format(self._BASE_URI, parameter))
            parameters = [Munch(parameter=parameter, value=value)]
        else:
            categories = _api.get(self._BASE_URI)
            parameters = []
            for category_name, category in categories.iteritems():
                parameters.extend(self._flatten_category({ category_name : category }))
            parameters.sort(key=sort_key)
        return parameters

    def _flatten_category(self, categories):
        for name, value in categories.iteritems():
            if isinstance(value, dict):
                for parameter in self._flatten_category(value):
                    parameter.parameter = "{}/{}".format(name, parameter.parameter)
                    yield parameter
            else:
                parameter = Munch(
                        parameter=name,
                        value=value,
                        )
                yield parameter

    @command(
        description="Set system configuration parameter",
        types=(ConfigParameter, Str),
        )
    @roles.INFINIDAT
    def set(self, _api, parameter, value):
        self._validate_parameter_value(parameter, value)
        old_value = _api.get("{}/{}".format(self._BASE_URI, parameter))
        if not isinstance(old_value, basestring):
            value_type = type(old_value)
            assert value_type in self._TYPES, "Unsupported parameter type: " + value_type
            value_type_name, converter = self._TYPES[value_type]
            converter = converter or value_type
            try:
                value = converter(value)
            except:
                raise BadValue('Cannot convert "{}" to {}'.format(value, value_type_name.lower()))
        _api.put("{}/{}".format(self._BASE_URI, parameter), value)
        return 'Changed configuration value "{}" from "{}" to "{}"'.format(parameter, old_value, value)

    def _validate_parameter_value(self, parameter, value):
        validator = self._VALIDATIONS.get(parameter)
        if validator:
            error = validator(value)
            if error is not None:
                raise BadValue('Cannot set "{}" to "{}": {}'.format(parameter, value, error))

@commandgroup(parent=ConfigCommandsBase, prefix="cache")
class CacheCommands(ConfigCommandsBase):

    @command(
        description="Set SSD cache for new pools",
        types=(YesNo,),
        substitute=dict(enabled="ssd"),
        )
    @roles.INFINIDAT.ADMIN
    def ssd_default(self, _api, enabled):
        return self._configure_param(_api,
            category="mgmt",
            param="pool.ssd_enabled_default",
            value=enabled,
            subject="SSD cache default for new pools",
            )
