from munch import Munch
from itertools import chain
from functools import partial
import time
import os
import ipaddress
from collections import defaultdict

from ..renderer import CollectionRenderer
from infi.shell import commandgroup, command, get_active_shell, render
from infi.shell.utils import strlist
from infi.shell.exceptions import BadArguments
from infi.shell.types import (
    Positive,
    Str,
    Int,
    IPAddress,
)



from ...utils import Version
from ..exceptions import InvalidSystemState, ConnectionFailed, ManagementException
from .types import (
    TimeDelta,
    MetricsSortByType,
    Pool,
    FilesystemOrClone,
    User,
    MetricsInterval,
    )
from ..system import system
from .options import QueryOptions
from .mapping import MappingCommandsBase
from . import EntityCommands, render_collection, set_renderer_query_fields, roles, LazyIdNameMapping
from .utils import getch_pause


class MetricsException(ManagementException):
    def __init__(self):
        error = Munch(
            code="METRICS_NOT_COLLECTED",
            message="The requested metrics could not be collected"
            )
        super(MetricsException, self).__init__(http_code="409",error_obj=error)


def get_filter(_api, protocol):
    try:
        response = _api.post("metrics/filters", dict(protocol=protocol))
    except ManagementException as e:
        # overide mgmt error since mgmt doesnt give any usefull information
        if e.code == "INVALID_FILTER_HANDLE_ID":
            raise MetricsException()
        raise e
    return response.id

def add_filter_to_filter(_api, filter_id, filter_field, filter_value):
    values = dict()
    values[filter_field] = filter_value
    response = _api.put("metrics/filters/%s" % filter_id, values)
    return response.id == filter_id

def get_collector(_api, collector_type, collected_fields, sorting_field=None, grouping_field=None, max_results=None,
                   interval=None, filters=None, filter_id=None):
    collector_dict = dict(
        filter_id=filter_id,
        filters=filters,
        type=collector_type,
        collected_fields= collected_fields,
        sorting_field=sorting_field,
        grouping_field=grouping_field,
        max_results=max_results,
        )
    collector_dict = {k: v for k, v in collector_dict.items() if v is not None}
    try:
        response = _api.post("metrics/collectors",collector_dict)
    except ManagementException as e:
        # overide mgmt error since mgmt doesnt give any usefull information
        if e.code == "INVALID_COLLECTOR_HANDLE_ID":
            raise MetricsException()
        raise e
    return response.id

def get_data_from_collector(_api, collector_ids):
    ids = "in:({})".format(",".join([str(x) for x in collector_ids]))
    response = _api.get('metrics/collectors/data', params=dict(collector_id=ids))
    modified_data = Munch()
    for collector in response.collectors:
        modified_data[collector.id] = collector
    return modified_data

def top_collector_data_to_table(_api, data):
    result_data = {}
    fields = data.fields
    if data.data == [] or data.data[0] == []:
        result_data = [Munch(zip(fields, ["-"] * 4))]
    else:
        result_data = [Munch(zip(fields, i)) for i in data.data[0] if len(i)]
    return result_data

def get_top_table(_api, collector_id, convert_data_function):
    data = get_data_from_collector(_api=_api, collector_ids=[collector_id])
    table =  top_collector_data_to_table(_api=_api, data=data[collector_id])
    result = convert_data_function(data=table) if convert_data_function else table
    return result

def add_element_names_to_data(_api, element_name, element_type, data):
    for i in data:
        if i[element_type] == '-':
            i["metrics_{}_name".format(element_name.__name__).lower()] = i[element_type]
            continue
        element = element_name.get_by_id(_api, i[element_type])
        i["metrics_{}_name".format(element_name.__name__).lower()] = element.name
    return data

def int_ip_to_long(element_type, data):
    for i in data:
        if i[element_type] == '-':
            continue
        ip = ipaddress.IPv4Address(int(i[element_type])).compressed
        ip = ip.split('.')
        ip.reverse()
        ip = '.'.join(ip)
        i[element_type] = ip
    return data

def ip_to_reverse_int(ip):
    ip = ip.split('.')
    ip.reverse()
    o = map(int, ip)
    res = (16777216 * o[0]) + (65536 * o[1]) + (256 * o[2]) + o[3]
    return res

# def get_show_filters(_api, protocol, element_types, element_ids, allow_metadata=False):
#     read_filter_id = get_filter(_api=_api, protocol='NFS')
#     add_filter_to_filter(_api=_api, filter_id=read_filter_id, filter_field="operation_type", filter_value="read")
#     write_filter_id = get_filter(_api=_api, protocol='NFS')
#     add_filter_to_filter(_api=_api, filter_id=write_filter_id, filter_field="operation_type", filter_value="write")
#     meta_filter_id = None
#     if allow_metadata:
#         meta_filter_id = get_filter(_api=_api, protocol='NFS')
#         add_filter_to_filter(_api=_api, filter_id=meta_filter_id, filter_field="operation_category", filter_value="meta data")
#     for element_type, element_id in zip(element_types, element_ids):
#         if element_id is None:
#             continue
#         add_filter_to_filter(_api=_api, filter_id=read_filter_id, filter_field=element_type, filter_value=element_id)
#         add_filter_to_filter(_api=_api, filter_id=write_filter_id, filter_field=element_type, filter_value=element_id)
#         if allow_metadata:
#             add_filter_to_filter(_api=_api, filter_id=meta_filter_id, filter_field=element_type, filter_value=element_id)
#     return read_filter_id, write_filter_id, meta_filter_id

def get_show_collectors(_api, collector_type, collected_fields, filters=None, filters_ids=None):
    read_collector = get_collector(_api=_api, filter_id=filters_ids[0] if filters_ids else None, filters=filters[0] if filters else None, collector_type=collector_type, collected_fields=collected_fields)
    write_collector = get_collector(_api=_api, filter_id=filters_ids[1] if filters_ids else None, filters=filters[1] if filters else None, collector_type=collector_type, collected_fields=collected_fields)
    meta_collector = None
    if (filters_ids and len(filters_ids) > 2) or (filters and len(filters) > 2):
        meta_collector = get_collector(_api=_api, filter_id=filters_ids[2] if filters_ids else None, filters=filters[2] if filters else None, collector_type=collector_type, collected_fields=collected_fields)
    return read_collector, write_collector, meta_collector

def get_show_table(_api, collector_ids):
    tmp_collctor_ids = [collector_ids[i] for i in collector_ids]
    data = get_data_from_collector(_api=_api, collector_ids=tmp_collctor_ids)
    data[collector_ids.read_collector]["metrics_collector_type"] = 'READ'
    data[collector_ids.write_collector]["metrics_collector_type"] = 'WRITE'
    if collector_ids.get("meta_collector", None):
        data[collector_ids.meta_collector]["metrics_collector_type"] = 'METADATA'
    result = []
    ############################## Code for showing reversed table : i.e read, write, meta are the header. Requires new implementation to renderer
    # for cat in categories:
    #     result[cat] = Munch(type=cat)
    # for collector in data:
    #     collector_data = data[collector]
    #     # data can be empty, in this case, fill zeros at all places
    #     if collector_data.data == []:
    #         for field in collector_data.fields:
    #             result[field][collector_data.metrics_collector_type] = 0
    #         continue
    #     for field, value in zip(collector_data.fields, collector_data.data[0]):
    #         result[field][collector_data.metrics_collector_type] = value
    ##############################
    for collector in data:
        collector_data = data[collector]
        if collector_data.data == []:
            for field in collector_data.fields:
                collector_data[field] = 0
        else:
            for field, value in zip(collector_data.fields, collector_data.data[0]):
                collector_data[field] = value
        result += [collector_data]
    return result



def run_loop(_api, loop_function, renderer_kwargs, interval=5, error_interval=10, count=None, message=None):
    output_stream = get_active_shell().output_stream
    render = partial(
        CollectionRenderer().render,
        stream=output_stream,
        header=True,
        **renderer_kwargs
        )

    context = Munch(
        print_data=loop_function(),
        render_flag=False,
        time_passed=0,
        )

    def loop(last_iteration=False, waiting=False):
        try:
            if context.print_data:
                # os.system('cls' if os.name == 'nt' else 'clear')
                print("\033[2J\033[1;1f")
                if message:
                    print(message + "\n\n")
                if waiting:
                    print("Collecting data...")
                else:
                    render(context.print_data, header=True)
            if last_iteration:
                return
            if getattr(getch_pause, 'prefer_Q', False):
                print('\n\nPress Q to return to the shell')
            getch_pause(interval)
            context.print_data = loop_function()
            context.time_passed = 0
        except (ConnectionFailed, ManagementException) as e:
            # overide mgmt error since mgmt doesnt know if this collector ever existed
            if e.code in ("INVALID_COLLECTOR_HANDLE_ID", "INVALID_FILTER_HANDLE_ID"):
                raise MetricsException()
            getch_pause(error_interval)
            context.time_passed += error_interval
            if context.time_passed > 500:
                raise e

    try:
        # first fetch should not be printed since its always 0
        loop(waiting=True)

        if count:
            for i in xrange(count):
                loop(last_iteration=i == count - 1)
            return

        while True:
            loop()
    except KeyboardInterrupt:
        pass


SYSTEM_QUERY_FIELDS = strlist("metrics_collector_type ops throughput latency")
SYSTEM_QUERY_SUBSTITUTES = dict(
    metrics_collector_type="---",
    ops="operations",
    )
SYSTEM_SHOW_RENDERER_KWARGS = dict(
        fields=SYSTEM_QUERY_FIELDS,
        substitute=SYSTEM_QUERY_SUBSTITUTES,
    )

POOL_QUERY_FIELDS = strlist("metrics_pool_name ops throughput latency")
POOL_QUERY_SUBSTITUTES = dict(
    metrics_pool_name="pool_name",
    ops="operations",
    )
POOL_TOP_RENDERER_KWARGS = dict(
        fields=POOL_QUERY_FIELDS,
        substitute=POOL_QUERY_SUBSTITUTES,
    )

FS_QUERY_FIELDS = strlist("metrics_filesystemorclone_name ops throughput latency")
FS_QUERY_SUBSTITUTES = dict(
    metrics_filesystemorclone_name="fs_name",
    ops="operations",
    )
FS_TOP_RENDERER_KWARGS = dict(
        fields=FS_QUERY_FIELDS,
        substitute=FS_QUERY_SUBSTITUTES,
    )

UID_QUERY_FIELDS = strlist("metrics_user_name user_id ops throughput latency")
UID_QUERY_SUBSTITUTES = dict(
    metrics_user_name="user_name",
    ops="operations",
    )
UID_TOP_RENDERER_KWARGS = dict(
        fields=UID_QUERY_FIELDS,
        substitute=UID_QUERY_SUBSTITUTES,
    )

GID_QUERY_FIELDS = strlist("metrics_group_name group_id ops throughput latency")
GID_QUERY_SUBSTITUTES = dict(
    metrics_group_name="group_name",
    ops="operations",
    )
GID_TOP_RENDERER_KWARGS = dict(
        fields=GID_QUERY_FIELDS,
        substitute=GID_QUERY_SUBSTITUTES,
    )

SOURCE_IP_QUERY_FIELDS = strlist("metrics_source_ip_name source_ip ops throughput latency")
SOURCE_IP_QUERY_SUBSTITUTES = dict(
    metrics_source_ip_name="source_ip",
    ops="operations",
    )
SOURCE_IP_TOP_RENDERER_KWARGS = dict(
        fields=SOURCE_IP_QUERY_FIELDS,
        substitute=SOURCE_IP_QUERY_SUBSTITUTES,
    )

DESTINATION_IP_QUERY_FIELDS = strlist("metrics_destination_ip_name destination_ip ops throughput latency")
DESTINATION_IP_QUERY_SUBSTITUTES = dict(
    metrics_destination_ip_name="destination_ip",
    ops="operations",
    )
DESTINATION_IP_TOP_RENDERER_KWARGS = dict(
        fields=DESTINATION_IP_QUERY_FIELDS,
        substitute=DESTINATION_IP_QUERY_SUBSTITUTES,
    )


@commandgroup(prefix="metrics")
class MetricsCommands(object):
    pass

@commandgroup(parent=MetricsCommands, prefix="nas")
class NasCommands(object):
    pass


@commandgroup(parent=NasCommands, prefix="system")
class NasSystemCommands(object):

    @command(
        description="Show metrics for system",
        types=(MetricsInterval, Positive),
        )
    @render(
        fields=SYSTEM_QUERY_FIELDS,
        substitute=SYSTEM_QUERY_SUBSTITUTES,
        )
    def show(self, _api, interval=5, count=1):
        filters = [
        dict(protocol="NFS",
            operation_type="read"
            ),
        dict(protocol="NFS",
            operation_type="write"
            ),
        dict(protocol="NFS",
            operation_category="meta data"
            ),
        ]
        read_collector, write_collector, meta_collector = get_show_collectors(_api=_api, filters_ids=None,filters=filters,
                                                                              collector_type="COUNTER", collected_fields=['ops','latency','throughput'])
        collectors_to_query = Munch(read_collector=read_collector, write_collector=write_collector)
        if meta_collector:
            collectors_to_query.meta_collector = meta_collector
        message = "System %s NAS performance" % system.name()
        run_loop(_api=_api, loop_function=partial(get_show_table, _api=_api, collector_ids=collectors_to_query), renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS,
                 interval=interval, error_interval=10, count=count, message=message)

@commandgroup(parent=NasCommands, prefix="pool")
class NasPoolCommands(object):

    ELEMENT = Pool
    ELEMENT_ID = "pool_id"

    @command(
        description="Show top metrics for pools in system",
        types=(MetricsSortByType, Positive),
        )
    def top(self, _api, sort_by=MetricsSortByType.TYPES['operations'], results=10):
        filter_id = get_filter(_api=_api, protocol='NFS')
        collector_id = get_collector(_api=_api, filter_id=filter_id, collector_type="TOP",  collected_fields=["throughput", "ops", "latency"], sorting_field=sort_by,
                                     grouping_field="pool_id", max_results=results)
        message = "Top pools in system '%s' By %s" % (system.name(), sort_by.upper())
        run_loop(_api=_api, loop_function=partial(get_top_table, _api=_api, collector_id=collector_id,
                 convert_data_function=partial(add_element_names_to_data, _api=_api, element_name=self.ELEMENT, element_type=self.ELEMENT_ID)),
                 renderer_kwargs=POOL_TOP_RENDERER_KWARGS, interval=1, error_interval=10, message=message)

    @command(
        description="Show metrics for specific pool in system",
        types=(Pool, MetricsInterval, Positive),
        substitute=dict(pool="name"),
        )
    def show(self, _api, pool, interval=5, count=1):
        filters = [
        dict(protocol="NFS",
            operation_type="read",
            pool_id=pool.id
            ),
        dict(protocol="NFS",
            operation_type="write",
            pool_id=pool.id
            ),
        dict(protocol="NFS",
            operation_category="meta data",
            pool_id=pool.id
            ),
        ]
        read_collector, write_collector, meta_collector = get_show_collectors(_api=_api, filters_ids=None,filters=filters,
                                                                              collector_type="COUNTER", collected_fields=['ops','latency','throughput'])
        collectors_to_query = Munch(read_collector=read_collector, write_collector=write_collector)
        if meta_collector:
            collectors_to_query.meta_collector = meta_collector
        message = "Pool %s NAS performance" % pool.name
        run_loop(_api=_api, loop_function=partial(get_show_table, _api=_api, collector_ids=collectors_to_query), renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS,
                 interval=interval, error_interval=10, count=count, message=message)


@commandgroup(parent=NasCommands, prefix="fs")
class NasFsCommands(object):

    ELEMENT = FilesystemOrClone
    ELEMENT_ID = "fs_entity_id"

    @command(
        description="Show top metrics for filesystems in system",
        types=(Pool, MetricsSortByType, Positive),
        )
    def top(self, _api,pool=None, sort_by=MetricsSortByType.TYPES['operations'], results=10):
        filter_id = get_filter(_api=_api, protocol='NFS')
        if pool:
            add_filter_to_filter(_api=_api, filter_id=filter_id, filter_field="pool_id", filter_value=pool.id)
        collector_id = get_collector(_api=_api, filter_id=filter_id, collector_type="TOP",  collected_fields=["throughput", "ops", "latency"], sorting_field=sort_by,
                                     grouping_field=self.ELEMENT_ID, max_results=results)

        message = "Top filesystems %s by %s" % ("for pool {}".format(pool.name) if pool else "in system", sort_by.upper())
        run_loop(_api=_api, loop_function=partial(get_top_table, _api=_api, collector_id=collector_id,
                 convert_data_function=partial(add_element_names_to_data, _api=_api, element_name=self.ELEMENT, element_type=self.ELEMENT_ID)),
                 renderer_kwargs=FS_TOP_RENDERER_KWARGS, interval=1, error_interval=10, message=message)

    @command(
        description="Show metrics for specific filesystem in system",
        types=(FilesystemOrClone, MetricsInterval, Positive),
        substitute=dict(fs="name"),
        )
    def show(self, _api, fs, interval=5, count=1):
        filters = [
        dict(protocol="NFS",
            operation_type="read",
            fs_entity_id=fs.id
            ),
        dict(protocol="NFS",
            operation_type="write",
            fs_entity_id=fs.id
            ),
        dict(protocol="NFS",
            operation_category="meta data",
            fs_entity_id=fs.id
            ),
        ]
        read_collector, write_collector, meta_collector = get_show_collectors(_api=_api, filters_ids=None,filters=filters,
                                                                              collector_type="COUNTER", collected_fields=['ops','latency','throughput'])
        collectors_to_query = Munch(read_collector=read_collector, write_collector=write_collector)
        if meta_collector:
            collectors_to_query.meta_collector = meta_collector

        message = "Filesystem %s NAS performance" % fs.name
        run_loop(_api=_api, loop_function=partial(get_show_table, _api=_api, collector_ids=collectors_to_query), renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS,
                 interval=interval, error_interval=10, count=count, message=message)



@commandgroup(parent=NasCommands, prefix="uid")
class NasUidCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "user_id"

    @command(
        description="Show top metrics for uid in system",
        types=(Pool, FilesystemOrClone, MetricsSortByType, Positive),
        )
    def top(self, _api, pool=None, fs=None, sort_by=MetricsSortByType.TYPES['operations'], results=10):
        filter_id = get_filter(_api=_api, protocol='NFS')
        if pool:
            add_filter_to_filter(_api=_api, filter_id=filter_id, filter_field="pool_id", filter_value=pool.id)
        if fs:
            add_filter_to_filter(_api=_api, filter_id=filter_id, filter_field="fs_entity_id", filter_value=fs.id)
        collector_id = get_collector(_api=_api, filter_id=filter_id, collector_type="TOP",  collected_fields=["throughput", "ops", "latency"], sorting_field=sort_by,
                                     grouping_field=self.ELEMENT_ID, max_results=results)

        if pool and fs:
            pre_message = "for pool %s, filesystem %s" % (pool.name, fs.name)
        elif pool:
            pre_message = "for pool %s" % (pool.name)
        elif fs:
            pre_message = "for filesystem %s" % (fs.name)
        else:
            pre_message = "in system"
        message = "Top UIDs %s by %s" % (pre_message, sort_by.upper())
        run_loop(_api=_api, loop_function=partial(get_top_table, _api=_api, collector_id=collector_id,
                 convert_data_function=None),
                 renderer_kwargs=UID_TOP_RENDERER_KWARGS, interval=1, error_interval=10, message=message)

    @command(
        description="Show metrics for specific uid in system",
        types=(Int, FilesystemOrClone, MetricsInterval, Positive),
        )
    def show(self, _api, uid, fs=None, interval=5, count=1):

        filters = [
        dict(protocol="NFS",
            operation_type="read",
            fs_entity_id=fs.id if fs else None,
            user_id=uid
            ),
        dict(protocol="NFS",
            operation_type="write",
            fs_entity_id=fs.id if fs else None,
            user_id=uid
            ),
        dict(protocol="NFS",
            operation_category="meta data",
            fs_entity_id=fs.id if fs else None,
            user_id=uid
            ),
        ]
        filters = [{k: v for k, v in collector_dict.items() if v is not None} for collector_dict in filters]
        read_collector, write_collector, meta_collector = get_show_collectors(_api=_api, filters_ids=None,filters=filters,
                                                                              collector_type="COUNTER", collected_fields=['ops','latency','throughput'])
        collectors_to_query = Munch(read_collector=read_collector, write_collector=write_collector)
        if meta_collector:
            collectors_to_query.meta_collector = meta_collector

        message = "UID %s NAS performance %s" % (uid, "for Filesystem {}".format(fs.name) if fs else "")
        run_loop(_api=_api, loop_function=partial(get_show_table, _api=_api, collector_ids=collectors_to_query), renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS,
                 interval=interval, error_interval=10, count=count, message=message)


@commandgroup(parent=NasCommands, prefix="gid")
class NasGidCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "group_id"

    @command(
        description="Show top metrics for gid in system",
        types=(Pool, FilesystemOrClone, MetricsSortByType, Positive),
        )
    def top(self, _api, pool=None, fs=None, sort_by=MetricsSortByType.TYPES['operations'], results=10):
        filter_id = get_filter(_api=_api, protocol='NFS')
        if pool:
            add_filter_to_filter(_api=_api, filter_id=filter_id, filter_field="pool_id", filter_value=pool.id)
        if fs:
            add_filter_to_filter(_api=_api, filter_id=filter_id, filter_field="fs_entity_id", filter_value=fs.id)
        collector_id = get_collector(_api=_api, filter_id=filter_id, collector_type="TOP",  collected_fields=["throughput", "ops", "latency"], sorting_field=sort_by,
                                     grouping_field=self.ELEMENT_ID, max_results=results)

        if pool and fs:
            pre_message = "for pool %s, filesystem %s" % (pool.name, fs.name)
        elif pool:
            pre_message = "for pool %s" % (pool.name)
        elif fs:
            pre_message = "for filesystem %s" % (fs.name)
        else:
            pre_message = "in system"
        message = "Top GIDs %s by %s" % (pre_message, sort_by.upper())
        run_loop(_api=_api, loop_function=partial(get_top_table, _api=_api, collector_id=collector_id,
                 convert_data_function=None),
                 renderer_kwargs=GID_TOP_RENDERER_KWARGS, interval=1, error_interval=10, message=message)

    @command(
        description="Show metrics for specific gid in system",
        types=(Int, FilesystemOrClone, MetricsInterval, Positive),
        )
    def show(self, _api, gid, fs=None, interval=5, count=1):
        filters = [
        dict(protocol="NFS",
            operation_type="read",
            fs_entity_id=fs.id if fs else None,
            group_id=gid
            ),
        dict(protocol="NFS",
            operation_type="write",
            fs_entity_id=fs.id if fs else None,
            group_id=gid
            ),
        dict(protocol="NFS",
            operation_category="meta data",
            fs_entity_id=fs.id if fs else None,
            group_id=gid
            ),
        ]
        filters = [{k: v for k, v in collector_dict.items() if v is not None} for collector_dict in filters]
        read_collector, write_collector, meta_collector = get_show_collectors(_api=_api, filters_ids=None,filters=filters,
                                                                              collector_type="COUNTER", collected_fields=['ops','latency','throughput'])
        collectors_to_query = Munch(read_collector=read_collector, write_collector=write_collector)
        if meta_collector:
            collectors_to_query.meta_collector = meta_collector
        message = "GID %s NAS performance %s" % (gid, "for Filesystem {}".format(fs.name) if fs else "")
        run_loop(_api=_api, loop_function=partial(get_show_table, _api=_api, collector_ids=collectors_to_query), renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS,
                 interval=interval, error_interval=10, count=count, message=message)


@commandgroup(parent=NasCommands, prefix="source_ip")
class NasSourceIPCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "source_ip"

    @command(
        description="Show top metrics for source ips in system",
        types=(Pool, FilesystemOrClone, MetricsSortByType, Positive),
        )
    def top(self, _api, pool=None, fs=None, sort_by=MetricsSortByType.TYPES['operations'], results=10):
        filter_id = get_filter(_api=_api, protocol='NFS')
        if pool:
            add_filter_to_filter(_api=_api, filter_id=filter_id, filter_field="pool_id", filter_value=pool.id)
        if fs:
            add_filter_to_filter(_api=_api, filter_id=filter_id, filter_field="fs_entity_id", filter_value=fs.id)
        collector_id = get_collector(_api=_api, filter_id=filter_id, collector_type="TOP",  collected_fields=["throughput", "ops", "latency"], sorting_field=sort_by,
                                     grouping_field=self.ELEMENT_ID, max_results=results)
        if pool and fs:
            pre_message = "for pool %s, filesystem %s" % (pool.name, fs.name)
        elif pool:
            pre_message = "for pool %s" % (pool.name)
        elif fs:
            pre_message = "for filesystem %s" % (fs.name)
        else:
            pre_message = "in system"
        message = "Top source IPs %s by %s" % (pre_message, sort_by.upper())
        run_loop(_api=_api, loop_function=partial(get_top_table, _api=_api, collector_id=collector_id,
                 convert_data_function=partial(int_ip_to_long, element_type=self.ELEMENT_ID)),
                 renderer_kwargs=SOURCE_IP_TOP_RENDERER_KWARGS, interval=1, error_interval=10, message=message)

    @command(
        description="Show metrics for specific source ip in system",
        types=(IPAddress, Pool, FilesystemOrClone, MetricsInterval, Positive),
        )
    def show(self, _api, source_ip, pool=None, fs=None, interval=5, count=1):

        filters = [
        dict(protocol="NFS",
            operation_type="read",
            fs_entity_id=fs.id if fs else None,
            pool_id=pool.id if pool else None,
            source_ip=ip_to_reverse_int(source_ip.ip)
            ),
        dict(protocol="NFS",
            operation_type="write",
            fs_entity_id=fs.id if fs else None,
            pool_id=pool.id if pool else None,
            source_ip=ip_to_reverse_int(source_ip.ip)
            ),
        dict(protocol="NFS",
            operation_category="meta data",
            fs_entity_id=fs.id if fs else None,
            pool_id=pool.id if pool else None,
            source_ip=ip_to_reverse_int(source_ip.ip)
            ),
        ]
        filters = [{k: v for k, v in collector_dict.items() if v is not None} for collector_dict in filters]
        read_collector, write_collector, meta_collector = get_show_collectors(_api=_api, filters_ids=None,filters=filters,
                                                                              collector_type="COUNTER", collected_fields=['ops','latency','throughput'])
        collectors_to_query = Munch(read_collector=read_collector, write_collector=write_collector)
        if meta_collector:
            collectors_to_query.meta_collector = meta_collector

        if pool and fs:
            pre_message = "for pool %s, filesystem %s" % (pool.name, fs.name)
        elif pool:
            pre_message = "for pool %s" % (pool.name)
        elif fs:
            pre_message = "for filesystem %s" % (fs.name)
        else:
            pre_message = "in system"
        message = "Source IP %s NAS performance %s" % (source_ip.ip, pre_message)
        run_loop(_api=_api, loop_function=partial(get_show_table, _api=_api, collector_ids=collectors_to_query), renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS,
                 interval=interval, error_interval=10, count=count, message=message)


@commandgroup(parent=NasCommands, prefix="destination_ip")
class NasDestinationIPCommands(object):

    ELEMENT = ''
    ELEMENT_ID = "destination_ip"

    @command(
        description="Show top metrics for destination ips in system",
        types=(Pool, FilesystemOrClone, MetricsSortByType, Positive),
        )
    def top(self, _api, pool=None, fs=None, sort_by=MetricsSortByType.TYPES['operations'], results=10):
        filter_id = get_filter(_api=_api, protocol='NFS')
        if pool:
            add_filter_to_filter(_api=_api, filter_id=filter_id, filter_field="pool_id", filter_value=pool.id)
        if fs:
            add_filter_to_filter(_api=_api, filter_id=filter_id, filter_field="fs_entity_id", filter_value=fs.id)
        collector_id = get_collector(_api=_api, filter_id=filter_id, collector_type="TOP",  collected_fields=["throughput", "ops", "latency"], sorting_field=sort_by,
                                     grouping_field=self.ELEMENT_ID, max_results=results)

        if pool and fs:
            pre_message = "for pool %s, filesystem %s" % (pool.name, fs.name)
        elif pool:
            pre_message = "for pool %s" % (pool.name)
        elif fs:
            pre_message = "for filesystem %s" % (fs.name)
        else:
            pre_message = "in system"
        message = "Top destination IPs %s by %s" % (pre_message, sort_by.upper())
        run_loop(_api=_api, loop_function=partial(get_top_table, _api=_api, collector_id=collector_id,
                 convert_data_function=partial(int_ip_to_long, element_type=self.ELEMENT_ID)),
                 renderer_kwargs=DESTINATION_IP_TOP_RENDERER_KWARGS, interval=1, error_interval=10, message=message)

    @command(
        description="Show metrics for specific destination ip in system",
        types=(IPAddress, Pool, FilesystemOrClone, MetricsInterval, Positive),
        substitute=dict(fs="name"),
        )
    def show(self, _api, destination_ip, pool=None, fs=None, interval=5, count=1):
        filters = [
        dict(protocol="NFS",
            operation_type="read",
            fs_entity_id=fs.id if fs else None,
            pool_id=pool.id if pool else None,
            destination_ip=ip_to_reverse_int(destination_ip.ip)
            ),
        dict(protocol="NFS",
            operation_type="write",
            fs_entity_id=fs.id if fs else None,
            pool_id=pool.id if pool else None,
            destination_ip=ip_to_reverse_int(destination_ip.ip)
            ),
        dict(protocol="NFS",
            operation_category="meta data",
            fs_entity_id=fs.id if fs else None,
            pool_id=pool.id if pool else None,
            destination_ip=ip_to_reverse_int(destination_ip.ip)
            ),
        ]
        filters = [{k: v for k, v in collector_dict.items() if v is not None} for collector_dict in filters]
        read_collector, write_collector, meta_collector = get_show_collectors(_api=_api, filters_ids=None,filters=filters,
                                                                              collector_type="COUNTER", collected_fields=['ops','latency','throughput'])
        collectors_to_query = Munch(read_collector=read_collector, write_collector=write_collector)
        if meta_collector:
            collectors_to_query.meta_collector = meta_collector

        if pool and fs:
            pre_message = "for pool %s, filesystem %s" % (pool.name, fs.name)
        elif pool:
            pre_message = "for pool %s" % (pool.name)
        elif fs:
            pre_message = "for filesystem %s" % (fs.name)
        else:
            pre_message = "in system"
        message = "Destination IP %s NAS performance %s" % (destination_ip.ip, pre_message)
        run_loop(_api=_api, loop_function=partial(get_show_table, _api=_api, collector_ids=collectors_to_query), renderer_kwargs=SYSTEM_SHOW_RENDERER_KWARGS,
                 interval=interval, error_interval=10, count=count, message=message)
