import uuid
from munch import Munch
from functools import partial
from itertools import izip

from infi.shell import commandgroup, render, runtime, get_active_shell
from infi.shell.types import YesNo, TimeDelta, Str, Hostname
from infi.shell.utils import strlist
from infi.shell.exceptions import BadArguments, CommandAborted, CommandFailed

from ...utils import (
    get_username_password,
    humanly_join,
    yesno,
    yesno_ensure_interactive,
    )
from ..remote import get_remote_system
from ..system import system
from ..exceptions import NoSuchEntity, ManagementException, ConnectionFailed
from ..renderer import InfiniboxRenderer
from .types import (
    Name,
    TimeDeltaOrNone,

    Replica,
    ReplicaRole,
    ReplicaDatasetPair,
    ReplicationLinkBySystemName,
    ReplicatedEntity,
    NonReplicatedEntity,
    NonReplicatedVolume,
    ReplicatedCg,
    NonReplicatedCg,
    NetworkSpace,

    Pool,
    Dataset,
    MasterVolume,
    VolumeSnap,
    ConsistencyGroup,
    )
from .options import QueryOptions
from .utils import is_detailed_query, increased_api_timeout
from . import (
    render_collection,
    roles,
    EntityCommands,
    LazyIdNameMapping,
    LazyEntityMapping,
    ENTITY,
    )

from ..pre_v2_patches import v2_command as command

def generate_entity_name():
    return str(uuid.uuid1())

class RemoteEntity(Str):
    def get_description(self):
        return "Name of a remote entity (auto-completion unavailable)"

class LazyIdNameMappingNoneDetached(LazyIdNameMapping):
    def _get_default(self, id):
        if id is None:
            return 'DETACHED'
        return LazyIdNameMapping._get_default(self, id)

class ReplicationCommandsMixin(object):
    def _should_retry_forcefully(self, exc):
        return exc.is_remote or exc.code in (
            "REMOTE_API_AUTHENTICATION_REQUIRED",
            "REPLICA_DELETION_ERROR",
            "REPLICA_DELETION_NOT_ALLOWED_ON_TARGET",
            "LINK_IO_ERROR",
            )

class ReplicaCommandsBase(EntityCommands, ReplicationCommandsMixin):
    ENTITY_TYPE = Replica

    def _result_message(self, entity, msg, subject=None, source=None):
        is_cg = entity.entity_type == "CONSISTENCY_GROUP"
        entity_type = ConsistencyGroup if is_cg else MasterVolume
        if not source:
            if is_cg:
                source = ConsistencyGroup.get_by_id(system.api, entity.local_cg_id)
            elif entity.get("entity_pairs"):
                pair = entity.entity_pairs[0]
                source = Munch(id=pair.local_entity_id, name=pair.local_entity_name)
            else:
                source = MasterVolume.get_by_id(system.api, entity.local_entity_id)
        return "Replica for {} {}".format(self._result_subject_and_entity(source, entity_type), msg)

    def _get_replica(self, _api, source, link=None):
        is_cg = "members_count" in source
        filter = _api.create_filter().eq("local_cg_id" if is_cg else "local_entity_id", source.id)
        if link:
            filter = filter.eq("link_id", link.id)
        replicas = Replica.get_all(_api, filter=filter)
        if len(replicas) != 1:
            template = "{} is not replicated{}" if not replicas else "{} has more than one replica{}"
            raise NoSuchEntity(template.format(
                self._result_subject_and_entity(source, ConsistencyGroup if is_cg else MasterVolume, capitalize=True),
                "" if not link else ' on "{}"'.format(link.remote_system_name)
                ))
        return replicas[0]

    def _create_replica_info(self, remote_api, link, source, interval, rpo, remote_pool=None, new_target_name=None):
        info = Munch(
            link_id=link.id,
            sync_interval=interval*1000 if interval != TimeDeltaOrNone.CONSTANTS.NONE else 0,
            rpo_value=rpo*1000 if rpo != TimeDeltaOrNone.CONSTANTS.NONE else 0,
            rpo_type="TIME",
            entity_pairs=[],
            )
        if remote_pool:
            info.remote_pool_id = Pool.get_by_name(remote_api, remote_pool).id
        if source._is_dataset:
            info.update(
                entity_type="VOLUME",
                )
        else:
            info.update(
                entity_type="CONSISTENCY_GROUP",
                local_cg_id=source.id,
                remote_cg_name=new_target_name,
                )
        return info

@commandgroup(prefix="replica")
class ReplicaCommands(ReplicaCommandsBase):
    QUERY_FIELDS = strlist("""
        _local_entity_name
        _entity_display_type
        role
        _remote_system_name
        _remote_entity_name
        state
        progress
        throughput
        rpo_state
        restore_point
        """)
    DETAILED_QUERY_FIELDS =strlist("""
        _local_entity_name
        _entity_display_type
        local_pool_name
        role
        _remote_system_name
        _remote_ip_addresses
        _remote_entity_name
        remote_pool_name
        state
        state_description
        progress
        throughput
        rpo_state
        rpo_value
        sync_interval
        created_at
        updated_at
        last_synchronized
        restore_point
        staging_area_allocated_size
        """)
    QUERY_SUBSTITUTES = dict(
        _remote_system_name="remote_system",
        _entity_display_type="type",
        _local_entity_name="local_dataset",
        _remote_entity_name="remote_dataset",
        _remote_ip_addresses="remote_ip_addresses",
        staging_area_allocated_size="staging_area_size",
        )
    SORTABLE_FIELDS = strlist("""
        role
        _local_entity_name
        _entity_display_type
        state
        restore_point
        """)

    @command(
        description="List existing replicas",
        types=(ReplicationLinkBySystemName, ReplicatedEntity.List, ReplicaRole),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, substitute=QUERY_SUBSTITUTES, sortable=SORTABLE_FIELDS, detailed=True),
        substitute=dict(
            link="system",
            datasets="local_dataset",
            ),
        )
    @render_collection(
        entity=Replica,
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        )
    def query(self, _api, link=None, datasets=None, role=None):
        link_mapping = LazyEntityMapping(ReplicationLinkBySystemName, _api, fields=("remote_system_name", "link_state"))

        # Workaround for staging area allocated size not being shown when querying multiple replicas
        prefetched_replica = None
        if datasets and len(datasets) == 1:
            prefetched_replica_filter = self._create_query_filter(_api, {
                "local_entity_id" if datasets[0]._is_dataset else "local_cg_id" : datasets[0].id,
                "link_id" : link,
                "role" : role
                })
            prefetched_replica = Replica.get_all(_api, filter=prefetched_replica_filter)
            if not prefetched_replica:
                return []
            if len(prefetched_replica) > 1:
                prefetched_replica = None
            else:
                prefetched_replica = [Replica.get_by_id(_api, prefetched_replica[0].id)]
        # End workaround

        return self._query(_api,
            prefetched_replica,
            link_id=link,
            role=role,
            _hook=partial(self._query_hook, _api=_api, lazy_link_by_id=link_mapping, datasets=datasets),
            _sort_on_client_side=True,
            # _default_sort=_api.create_sort().asc("role").asc("_local_entity_name"),
            )

    def _query_hook(self, replicas, _api=None, lazy_link_by_id=None, datasets=None):
        dataset_ids = set((dataset.id for dataset in datasets)) if datasets else None
        for r in replicas:
            is_cg_replica = r.entity_type == "CONSISTENCY_GROUP"
            if datasets:
                if is_cg_replica and r.local_cg_id not in dataset_ids \
                or not is_cg_replica and not any(pair.local_entity_id in dataset_ids for pair in r.entity_pairs):
                    continue
            r = Munch((k, v) for k, v in r.iteritems() if k[0] != "_")
            if lazy_link_by_id:
                link = lazy_link_by_id[r.link_id]
                r._remote_system_name = link.remote_system_name
                if link.link_state == "DOWN":
                    r.state_reason = "LINK_DOWN"
            r._entity_display_type = "CG" if is_cg_replica else r.entity_type
            r._local_entity_name, r._remote_entity_name = [], []
            if is_cg_replica:
                r._local_entity_name = r.get("local_cg_name") or "<Local CG: {}>".format(r.local_cg_id)
                r._remote_entity_name = r.get("remote_cg_name") or "<Remote CG: {}>".format(r.remote_cg_id)
            else:
                r._local_entity_name = r.entity_pairs[0].local_entity_name if r.entity_pairs else "N/A"
                r._remote_entity_name = r.entity_pairs[0].remote_entity_name if r.entity_pairs else "N/A"
            if r.role == "TARGET":
                r.update(
                    state="TARGET",
                    progress="N/A",
                    throughput="N/A",
                    )
            else:
                r.progress = "{}%".format(r['progress']) if r['progress'] is not None else None
            yield r


    CG_MEMBER_QUERY_FIELDS = strlist("local_dataset remote_dataset")

    @command(
        description="List members of replicated local and remote consistency groups",
        types=(ReplicatedCg,),
        options=QueryOptions(fields=CG_MEMBER_QUERY_FIELDS, sortable=CG_MEMBER_QUERY_FIELDS),
        substitute=dict(
            link="system",
            cg="local_cg",
            ),
        )
    @render(
        fields=CG_MEMBER_QUERY_FIELDS,
        )
    def member_query(self, _api, cg):
        return self._query(_api,
            local_cg_id=cg.id,
            _hook=self._cg_member_query_hook,
            _sort_on_client_side=True,
            _default_sort=_api.create_sort().asc("local_dataset"),
            )

    def _cg_member_query_hook(self, replicas):
        for replica in replicas:
            for pair in replica.entity_pairs or ():
                yield Munch(
                    local_dataset=pair.local_entity_name,
                    remote_dataset=pair.remote_entity_name,
                    )

    @command(
        description="Create a new replica along with corresponding remote datasets",
        types=(ReplicationLinkBySystemName, NonReplicatedEntity, TimeDeltaOrNone, TimeDeltaOrNone, RemoteEntity, Name, Name),
        substitute=dict(
            link="system",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, link, source, interval, rpo, remote_pool, new_target_name=None, new_datasets_suffix=None):
        if source._is_dataset and new_datasets_suffix:
            raise BadArguments("new_datasets_suffix is only applicable when source is a consistency group")
        with get_remote_system(_api, link).api as remote_api:
            replica_info = self._create_replica_info(
                remote_api=remote_api,
                link=link,
                source=source,
                interval=interval,
                rpo=rpo,
                remote_pool=remote_pool,
                new_target_name=new_target_name,
                )
            if source._is_dataset:
                replica_info.entity_pairs.append(Munch(
                    local_entity_id=source.id,
                    local_base_action="NO_BASE_DATA",
                    remote_base_action="CREATE",
                    remote_entity_name=new_target_name,
                    ))
            else:
                replica_info.entity_pairs.extend(Munch(
                    local_entity_id=member.id,
                    local_base_action="NO_BASE_DATA",
                    remote_base_action="CREATE",
                    remote_entity_name="{}{}".format(member.name, new_datasets_suffix or ""),
                    ) for member in Dataset.iter_all(_api, filter=_api.create_filter(cg_id=source.id)))
            return self._create(_api, **replica_info)

    @command(
        description="Delete replica",
        types=(ReplicatedEntity, YesNo, YesNo),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def delete(self, _api, source, retain_staging_area=False, force=False):
        replica = self._get_replica(_api, source)
        params = dict(
            retain_staging_area=retain_staging_area,
            )
        send_request = partial(self._del, _api, replica, params=params)
        result = None
        try:
            with get_remote_system(_api, replica).api:
                result = send_request()
        except ManagementException as e:
            if not self._should_retry_forcefully(e):
                raise
            elif not force:
                if not yesno("{}{}Would you like to perform the operation forcefully?".format(e.message,"\n" if e.message else ""), always_prompt=True):
                    raise CommandAborted
            params.update(
                force_if_no_remote_credentials="true",
                force_if_remote_error="true",
                force_on_target="true",
                )
            result = send_request(approved=True)
        staging_message = "and its staging area " + ("retained" if retain_staging_area else "discarded")
        if retain_staging_area and result and result.entity_pairs:
            non_reclaimed_names = sorted(pair.local_entity_name
                                         for pair in result.entity_pairs
                                         if not pair.get("_local_reclaimed_snapshot_id", False))
            if non_reclaimed_names:
                staging_message = "but there was nothing in the staging area to retain" if len(result.entity_pairs) == len(non_reclaimed_names) \
                                  else "and its staging area was retained, except for the following datasets: " + humanly_join(non_reclaimed_names)
        return self._result_message(replica, "deleted, " + staging_message, source=source)

    @command(
        description="Set replica sync interval",
        types=(ReplicatedEntity, TimeDelta),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_interval(self, _api, source, interval):
        replica = self._get_replica(_api, source)
        with get_remote_system(_api, replica).api:
            self._update(_api, replica, sync_interval=interval*1000)
        return self._result_message(replica, "set to synchronize every {} seconds".format(interval), source=source)

    @command(
        description="Set replica recovery point objective (RPO)",
        types=(ReplicatedEntity, TimeDelta),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def set_rpo(self, _api, source, rpo):
        replica = self._get_replica(_api, source)
        if replica.rpo_value == rpo*1000:
            return self._result_message(replica, "recovery point objective already set to {} seconds".format(rpo), source=source)
        with get_remote_system(_api, replica).api:
            self._update(_api, replica,
                rpo_value=rpo*1000,
                )
        return self._result_message(replica, "recovery point objective set to {} seconds".format(rpo), source=source)

    @command(
        description="Suspend replica",
        types=(ReplicatedEntity,),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def suspend(self, _api, source):
        replica = self._get_replica(_api, source)
        with get_remote_system(_api, replica).api:
            self._post(_api, replica, "suspend")
        return self._result_message(replica, "suspended", source=source)

    @command(
        description="Start a sync job of a replica immediately",
        types=(ReplicatedEntity,),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def sync_now(self, _api, source):
        replica = self._get_replica(_api, source)
        with get_remote_system(_api, replica).api:
            self._post(_api, replica, "sync")
        return self._result_message(replica, "sync started", source=source)

    @command(
        description="Resume replica",
        types=(ReplicatedEntity, YesNo),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def resume(self, _api, source=None, resume_all=None):
        if source and resume_all:
            raise BadArguments, "local_dataset and resume_all parameters are mutually exclusive"
        if source is None and resume_all is None:
            raise BadArguments, "Please use either local_dataset or resume_all parameters"
        if source:
            replica = self._get_replica(_api, source)
            with get_remote_system(_api, replica).api:
                with increased_api_timeout():
                    self._post(_api, replica, "resume")
            return self._result_message(replica, "resumed", source=source)
        if resume_all:
            active_shell = get_active_shell()
            links = dict()
            for replica in Replica.iter_all(_api):
                if replica.role == "TARGET":
                        continue
                links.setdefault(replica.link_id,[]).append(replica)
            for link in links.iteritems():
                with get_remote_system(_api, link[0]).api as remote_api:
                    try:
                        remote_api.get("system/version")
                    except:
                        continue
            for _, items in links.iteritems():
                for replica in items:
                    entity = ("volume %s" % replica.entity_pairs[0].local_volume.name
                    if replica.entity_type == "VOLUME"
                    else "consistency group %s" % replica.local_cg_name)
                    with get_remote_system(_api, replica).api:
                        try:
                            print '--------------------\n %s' % entity
                            self._post(_api, replica, "resume")
                            active_shell.output("--- Replica resumed")
                        except Exception as e:
                            active_shell.error("--- Replica for %s: %s" % (entity, e))
        if resume_all is False:
            raise BadArguments, 'local_dataset parameter must be used'




    @command(
        description="Change a replica's role one-sidedly",
        types=(ReplicatedEntity,),
        substitute=dict(
            source="local_dataset",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def change_role(self, _api, source):
        replica = self._get_replica(_api, source)
        with get_remote_system(_api, replica).api:
            replica = self._post(_api, replica, "change_role")
        return self._result_message(replica, "role changed to {}".format(replica.role), source=source)

@commandgroup(parent=ReplicaCommands, prefix="advanced")
class ReplicaAdvancedCommands(ReplicaCommandsBase):

    @command(
        description="Create a new replica to target dataset that already exists on remote system",
        types=(ReplicationLinkBySystemName, NonReplicatedVolume, TimeDeltaOrNone, TimeDelta, RemoteEntity, VolumeSnap, RemoteEntity),
        substitute=dict(
            link="system",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create(self, _api, link, source, interval, rpo, target, source_snapshot=None, target_snapshot=None):
        if source_snapshot:
            if not target_snapshot and not system.user.roles.is_infinidat():
                raise BadArguments("Please specify source_snapshot and target_snapshot, or neither")
        with get_remote_system(_api, link).api as remote_api:
            replica_info = self._create_replica_info(
                remote_api=remote_api,
                link=link,
                source=source,
                interval=interval,
                rpo=rpo,
                )
            pair = Munch(
                local_entity_id=source.id,
                remote_entity_id=MasterVolume.get_by_name(remote_api, target).id,
                local_base_action="NO_BASE_DATA",
                remote_base_action="NO_BASE_DATA",
                )
            if source_snapshot and target_snapshot:
                pair.update(
                    local_base_action="BASE",
                    remote_base_action="BASE",
                    local_base_entity_id=source_snapshot.id,
                    remote_base_entity_id=VolumeSnap.get_by_name(remote_api, target_snapshot).id,
                    )
            replica_info.entity_pairs.append(pair)
            success_message = 'Replica from volume "{}" to volume "{}" created'.format(source.name, target)
            return self._create(_api, returns=success_message, **replica_info)

    @command(
        description="Create a new replica to target consistency group that already exists on remote system",
        types=(ReplicationLinkBySystemName, NonReplicatedCg, TimeDeltaOrNone, TimeDelta, RemoteEntity, ReplicaDatasetPair.List),
        substitute=dict(
            link="system",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def create_cg(self, _api, link, source, interval, rpo, target, cg_pairs=None):
        with get_remote_system(_api, link).api as remote_api:
            replica_info = self._create_replica_info(
                remote_api=remote_api,
                link=link,
                source=source,
                interval=interval,
                rpo=rpo,
                )
            remote_cg = ConsistencyGroup.get_by_name(remote_api, target)
            replica_info.update(
                remote_cg_id=remote_cg.id,
                )
            local_members = Dataset.get_all(_api, filter=_api.create_filter(cg_id=source.id))
            remote_members = Dataset.get_all(remote_api, filter=_api.create_filter(cg_id=remote_cg.id))
            if len(local_members) != len(remote_members):
                raise CommandFailed('Invalid consistency groups: Local group "{}" has {} members, but remote group "{}" has {} members'
                                    .format(source.name, len(local_members), remote_cg.name, len(remote_members)))

            def add_pair(local_member, remote_member):
                replica_info.entity_pairs.append(Munch(
                    local_entity_id=local_member.id,
                    local_base_action="NO_BASE_DATA",
                    remote_base_action="NO_BASE_DATA",
                    remote_entity_id=remote_member.id,
                    ))

            def process_member(all_members, member_name, processed_set, locality, cg):
                if member_name in processed_set:
                    raise BadArguments('{} dataset "{}" was specified more than once in cg_pairs'.format(locality, member_name))
                member = all_members.pop(member_name, None)
                if not member:
                    raise NoSuchEntity('No such dataset: {} consistency group "{}" has no such member: "{}"'
                                       .format(locality, cg.name, member_name))
                processed_set.add(member_name)
                return member

            if cg_pairs:  # Manual mapping
                processed_local_members, processed_remote_members = set(), set()
                local_members_by_name = {dataset.name : dataset for dataset in local_members}
                remote_members_by_name = {dataset.name : dataset for dataset in remote_members}
                for local_member_name, remote_member_name in cg_pairs:
                    add_pair(
                        process_member(local_members_by_name, local_member_name, processed_local_members, "Local", source),
                        process_member(remote_members_by_name, remote_member_name, processed_remote_members, "Remote", target),
                        )
            else:  # Automatic mapping
                sort_members = partial(sorted, key=lambda m: (m.size, m.name))
                for local_member, remote_member in izip(sort_members(local_members), sort_members(remote_members)):
                    if local_member.size != remote_member.size:
                        raise CommandFailed("Invalid consistency groups: Sizes of local datasets do not match sizes of remote datasets")
                    add_pair(local_member, remote_member)
                if replica_info.entity_pairs and not self._get_user_confirmation_for_create_cg(replica_info.entity_pairs, local_members, remote_members):
                    raise CommandAborted("If you wish to specify the configuration manually, use cg_pairs")

            replica = self._create(_api, returns=ENTITY, **replica_info)
            return self._result_message(replica, "created" if cg_pairs else "created using automatic configuration", source=source)

    def _get_user_confirmation_for_create_cg(self, entity_pairs, local_members, remote_members):
        if yesno_ensure_interactive():
            # User should not be prompted
            return True
        local_members_by_id = {dataset.id : dataset for dataset in local_members}
        remote_members_by_id = {dataset.id : dataset for dataset in remote_members}
        InfiniboxRenderer().render(
            (Munch(
                local_dataset=local_members_by_id[p.local_entity_id].name,
                local_size=local_members_by_id[p.local_entity_id].size,
                remote_dataset=remote_members_by_id[p.remote_entity_id].name,
                remote_size=remote_members_by_id[p.remote_entity_id].size,
                ) for p in entity_pairs),
            stream=get_active_shell().output_stream,
            fields=strlist("local_dataset local_size remote_dataset remote_size"),
            )
        return yesno("Please examine the table above. Is this the configuration you intended for?")

@commandgroup(prefix="link")
class ReplicationLinkCommands(EntityCommands, ReplicationCommandsMixin):
    ENTITY_TYPE = ReplicationLinkBySystemName
    QUERY_FIELDS = strlist("remote_system_name link_state last_connection_timestamp _local_network_space remote_host")
    DETAILED_QUERY_FIELDS = QUERY_FIELDS + strlist(" _remote_network_space remote_system_serial_number")
    INFINIDAT_DETAILED_QUERY_FIELDS = DETAILED_QUERY_FIELDS + strlist("name")
    QUERY_SUBSTITUTES = dict(
        remote_system_name="remote_system",
        last_connection_timestamp="last_connection",
        remote_host="remote_system_address",
        remote_system_serial_number="remote_system_serial",
        _local_network_space="local_network_space",
        _remote_network_space="remote_network_space",
        name="link_name",
        )

    @command(
        description="List existing replication links",
        types=(ReplicationLinkBySystemName.List, NetworkSpace),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, infinidat_fields=INFINIDAT_DETAILED_QUERY_FIELDS, sortable=QUERY_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        substitute=dict(
            links="remote_system",
            ),
        )
    @render(
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        detailed_individually=True,
        detailed_fields=DETAILED_QUERY_FIELDS,
        alignment=dict(
            _local_network_space="left",
            _remote_network_space="left",
            ),
        infinidat=dict(
            detailed_individually=True,
            detailed_fields=INFINIDAT_DETAILED_QUERY_FIELDS,
            ),
        )
    def query(self, _api, links=None, net_space=None):
        netspace_mapping = {net_space.id : net_space.name} if net_space \
                           else LazyIdNameMappingNoneDetached(NetworkSpace, _api)
        return self._query(_api, links,
            local_replication_network_space_id=net_space,
            _hook=partial(self._query_hook, api=_api, lazy_network_spaces_by_id=netspace_mapping),
            )

    def _query_hook(self, links, api=None, lazy_network_spaces_by_id=None):
        should_get_remote_netspace = api and (is_detailed_query() or {"_remote_network_space", "*"}.intersection(runtime.options.get("columns") or ()))
        if not lazy_network_spaces_by_id and not should_get_remote_netspace:
            return links
        for link in links:
            if lazy_network_spaces_by_id:
                link._local_network_space = lazy_network_spaces_by_id[link.local_replication_network_space_id]
            if should_get_remote_netspace:
                link._remote_network_space = None
                with get_remote_system(api, link).api as remote_api:
                    try:
                        remote_netspace = remote_api.get(NetworkSpace.get_uri(link.remote_replication_network_space_id), handle_remote_auth_error=False, timeout=5)
                        link._remote_network_space = remote_netspace.name
                    except ManagementException as e:
                        reason = "authentication required" if e.code == api.MGMT_ERROR_CODES.REMOTE_AUTH else e.code
                        link._remote_network_space = "N/A ({})".format(reason) if reason else "N/A"
                    except ConnectionFailed as e:
                        link._remote_network_space = "N/A (timeout)"
        return links

    @command(
        description="Create a new replication link",
        types=(Hostname, NetworkSpace),
        substitute=dict(
            address="remote_address",
            ),
        )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, address, network_space):
        name = generate_entity_name()
        with _api.remote_auth_context(address) as ctx:
            link = self._create_link(_api, name, address, network_space)
            if ctx.username and ctx.password:
                get_remote_system(_api, link).api.set_basic_auth(ctx.username, ctx.password)
        return self._result_message(link, "created")

    def _create_link(self, _api, name, address, network_space):
        return self._create(_api,
            name=name,
            remote_host=address,
            local_replication_network_space_id=network_space.id,
            returns=ENTITY,
            )

    @command(
        description="Delete replication link",
        types=(ReplicationLinkBySystemName, YesNo),
        substitute=dict(
            link="remote_system",
            ),
        )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, link, force=False):
        send_request = partial(self._del, _api, link)
        try:
            with get_remote_system(_api, link).api:
                send_request()
        except ManagementException as e:
            if not self._should_retry_forcefully(e):
                raise
            elif not force:
                if not yesno("{}{}Would you like to perform the operation forcefully?".format(e.message,"\n" if e.message else ""), always_prompt=True):
                    raise CommandAborted
            send_request(approved=True, params=dict(
                force_if_no_remote_credentials="true",
                force_if_remote_error="true",
                ))
        return self._result_message(link, "deleted")

    @command(
        description="Authenticate on remote system",
        types=(ReplicationLinkBySystemName, Str, Str),
        substitute=dict(
            link="remote_system",
            ),
        )
    @roles.INFINIDAT.ADMIN.POOLADMIN
    def authenticate(self, _api, link, username=None, password=None):
        if password and not username:
            raise BadArguments("When specifying password, username must be specified")
        if not password:
            username, password = get_username_password(
                username=username,
                prompt="Username for {}".format(link.remote_system_name),
                )
        remote_system = get_remote_system(_api, link)
        remote_system.login(username, password)
        return 'Authenticated on {} as user "{}" with role {}'.format(
            link.remote_system_name,
            remote_system.user.username,
            ", ".join(remote_system.user.roles.roles),
            )

    @command(description='Attach network space to link',
             types=(ReplicationLinkBySystemName, NetworkSpace),
             substitute={'link': 'remote_system'})
    @roles.INFINIDAT.ADMIN
    def attach(self, _api, link, network_space):
        with increased_api_timeout():
            self._put(_api, link, data={
                'local_replication_network_space_id': network_space.id,
            })
        print('Link that uses network space "%s" is attached to the remote system %s' %
                (network_space.name, self._result_entity_name(link)))
        if not yesno("Link is attached. Do you want to refresh now to synchronize configuration to the remote system?"):
            return self._result_message(link, "was not refreshed.", "\n")
        return self.refresh(_api=_api, link=link)

    @command(description='Detach network space from link',
             types=(ReplicationLinkBySystemName,),
             substitute={'link': 'remote_system'})
    @roles.INFINIDAT.ADMIN
    def detach(self, _api, link):
        with increased_api_timeout():
            self._put(_api, link, data={
                'local_replication_network_space_id': None,
            })
        return 'Link is detached from the remote system %s' % self._result_entity_name(link)

    @command(description='Refresh link',
             types=(ReplicationLinkBySystemName,),
             substitute={'link': 'remote_system'})
    @roles.INFINIDAT.ADMIN
    def refresh(self, _api, link):
        # The operation requires authentication on the remote system.
        send_request = partial(self._post, _api, link, 'refresh', data={})
        with get_remote_system(_api, link).api as remote_api:
            with increased_api_timeout():
                send_request()

        return self._result_message(link, 'refreshed')

    @command(description='Change remote address',
             types=(ReplicationLinkBySystemName, Hostname),
             substitute={'link': 'remote_system', 'address': 'remote_address'})
    @roles.INFINIDAT.ADMIN
    def change_remote_address(self, _api, link, address):
        self._put(_api, link, data={'remote_host': address})
        return self._result_message(link, 'remote address changed')
