from functools import partial

from infi.shell.commands import commandgroup, command, render
from infi.shell.utils import strlist

from .types import Drive, Enclosure
from .options import QueryOptions
from . import EntityCommands, roles

QUERY_FIELDS           = strlist("name state              bytes_capacity                                                                state_description")
QUERY_INFINIDAT_FIELDS = strlist("name state connectivity bytes_capacity vendor model product_rev serial_number firmware failure_reason state_description")
QUERY_SUBSTITUTES = dict(
    bytes_capacity="capacity",
    product_rev="revision",
    )

@commandgroup(prefix="drive")
class DriveCommands(EntityCommands):
    ENTITY_TYPE = Drive

    @command(
        description="List all data-drives",
        types=(Drive.List, Enclosure.List),
        substitute=dict(
            drive="name",
            ),
        options=QueryOptions(fields=QUERY_FIELDS, infinidat_fields=QUERY_INFINIDAT_FIELDS, substitute=QUERY_SUBSTITUTES),
        )
    @render(
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        infinidat=dict(
            fields=QUERY_INFINIDAT_FIELDS,
            )
        )
    def query(self, _api, drive=None, enclosure=None):
        enclosure_ids = set(e.id for e in enclosure) if enclosure else None
        return self._query(_api, drive, _hook=partial(self._query_hook, enclosure_ids=enclosure_ids))

    def _query_hook(self, drives, enclosure_ids=None):
        for drive in drives:
            if enclosure_ids and drive.enclosure_index not in enclosure_ids:
                continue
            drive.bytes_capacity = drive.bytes_capacity or None
            drive.connectivity = " ".join("N{}".format(node_index) if connected else "--" for node_index, connected in enumerate(drive.nodes_access, 1))
            if drive.get("failure_reason") == "none":
                drive.failure_reason = None
            yield drive

    @command(
        description="Initiate drive activation process",
        types=(Drive,),
        substitute=dict(
            drive="name",
            ),
        )
    @roles.INFINIDAT.TECHNICIAN
    def activate(self, _api, drive):
        self._drive_action(_api, drive, "phase_in")
        return "Activating drive {} of enclosure {}".format(drive.drive_index, drive.enclosure_index)

    @command(
        description="Initiate drive deactivation process",
        types=(Drive,),
        substitute=dict(
            drive="name",
            ),
        )
    @roles.INFINIDAT.TECHNICIAN
    def deactivate(self, _api, drive):
        self._drive_action(_api, drive, "phase_out")
        return "Deactivating drive {} of enclosure {}".format(drive.drive_index, drive.enclosure_index)

    @command(
        description="Initiate drive testing process (required before activation is allowed)",
        types=(Drive,),
        substitute=dict(
            drive="name",
            ),
        )
    @roles.INFINIDAT.TECHNICIAN
    def test(self, _api, drive):
        self._drive_action(_api, drive, "test")
        return "Tested drive {} of enclosure {}".format(drive.drive_index, drive.enclosure_index)

    def _drive_action(self, _api, drive, state):
        _api.post("components/enclosures/{}/drives/{}/{}".format(drive.enclosure_index, drive.drive_index, state))
