from itertools import chain
from functools import partial

from infi.shell import commandgroup, command as _command, render
from infi.shell.types import YesNo
from infi.shell.utils import strlist, NOTHING

from ..types import (
    Name,
    FCPort,
    FCPortRole,
    FCSwitch,
    FCSoftTarget,
    FCSoftTargetCreationQuantity,
    Node,
    )
from ..options import QueryOptions
from ..utils import get_patch_panel_ports
from .. import EntityCommands, roles, LazyIdNameMapping, LazyEntityMapping
from . import ConfigCommandsBase

from ...pre_v2_patches import v2_command as command

@commandgroup(parent=ConfigCommandsBase, prefix="fc")
class FCCommands(ConfigCommandsBase):
    pass

@commandgroup(parent=FCCommands, prefix="port")
class FCPortCommands(ConfigCommandsBase, EntityCommands):
    ENTITY_TYPE = FCPort
    QUERY_FIELDS           = strlist("name enabled state state_description link_state speed _switch_name wwpn role                       connection_speed max_speed patch_panel_label")
    INFINIDAT_QUERY_FIELDS = strlist("name enabled state state_description link_state speed _switch_name wwpn role vendor model firmware connection_speed max_speed patch_panel_label")
    QUERY_SUBSTITUTE = dict(
        link_state="link",
        _switch_name="switch",
        connection_speed="speed",
        patch_panel_label="patch_panel",
        )

    @_command(
        description="List system FC ports",
        types=(FCPort.List, FCSwitch, Node),
        substitute=dict(
            ports="name",
            ),
        options=QueryOptions(fields=QUERY_FIELDS, infinidat_fields=INFINIDAT_QUERY_FIELDS, substitute=QUERY_SUBSTITUTE),
        )
    @render(
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTE,
        infinidat=dict(
            fields=INFINIDAT_QUERY_FIELDS,
            ),
        technician=dict(
            fields=INFINIDAT_QUERY_FIELDS,
            ),
        )
    def query(self, _api, ports=None, switch=None, node=None):
        lazy_switch_names_by_wwnn = LazyEntityMapping(FCSwitch, _api, fields=["name"], mapping_key="wwnn")
        patch_panel_ports = get_patch_panel_ports("FC")
        return self._query(_api, ports,
            switch_wwnn=switch.wwnn if switch else None,
            node_index=node,
            _hook=partial(self._query_hook,
                lazy_switch_names_by_wwnn=lazy_switch_names_by_wwnn,
                patch_panel_ports=patch_panel_ports,
                ),
            )

    def _query_hook(self, ports, lazy_switch_names_by_wwnn=None, patch_panel_ports=None):
        for port in ports:
            port._switch_name = None
            if lazy_switch_names_by_wwnn and port.switch_wwnn:
                port._switch_name = lazy_switch_names_by_wwnn[port.switch_wwnn].name
            port.patch_panel_label = None
            pp_port = None if not patch_panel_ports else patch_panel_ports.get((port.node_index, port.hba_port_number))
            if pp_port:
                port.patch_panel_label = (pp_port.label or pp_port.port_num)
        return ports

    def _get_port_url(self, port, *parts):
        return "/".join(chain(["components/nodes/{p.node_index}/fc_ports/{p.id}".format(p=port)], parts))

    @command(
        name="enable",
        description="Enable FC port",
        types=(FCPort, FCPortRole),
        substitute=dict(
            port="name",
            ),
        )
    @roles.INFINIDAT.TECHNICIAN
    def infinidat__enable(self, _api, port, role=None):
        data = dict(role=role) if role else NOTHING
        _api.post(self._get_port_url(port, "enable"), data=data)
        return self._result_message(port, "enabled" if role is None
            else "enabled, and its role {} {}".format("remained" if role == port.role else "changed to", role))

    @command(
        description="Enable FC port",
        types=(FCPort,),
        substitute=dict(
            port="name",
            ),
        )
    @roles.ADMIN
    def enable(self, _api, port):
        _api.post(self._get_port_url(port, "enable"))
        return self._result_message(port, "enabled")

    @command(
        description="Disable FC port",
        types=(FCPort,),
        substitute=dict(
            port="name",
            ),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def disable(self, _api, port):
        _api.post(self._get_port_url(port, "disable"))
        return self._result_message(port, "disabled")

@commandgroup(parent=FCCommands, prefix="switch")
class FCSwitchCommands(EntityCommands):
    ENTITY_TYPE = FCSwitch
    QUERY_FIELDS = strlist("name wwnn resiliency_state vendor")
    QUERY_SUBSTITUTE = dict(
        resiliency_state="resiliency",
        )
    SORTABLE_FIELDS = strlist("name wwn vendor")

    @command(
        description="List FC switches",
        types=(FCSwitch.List,),
        substitute=dict(
            switches="name",
            ),
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS),
        )
    @render(
        fields=QUERY_FIELDS,
        )
    def query(self, _api, switches=None):
        return self._query(_api, switches)

    @command(
        description="Rename FC switch",
        types=(FCSwitch, Name),
        substitute=dict(
            switch="name",
            ),
        )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, switch, new_name):
        return self._rename(_api, switch, new_name)

    @command(
        description="Delete FC switch",
        types=(FCSwitch.List,),
        substitute=dict(
            switches="name",
            ),
        )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, switches):
        return self._delete(_api, switches)

@commandgroup(parent=FCCommands, prefix="soft_target")
class FCSoftTargetCommands(EntityCommands):
    ENTITY_TYPE = FCSoftTarget
    QUERY_FIELDS = strlist("wwpn port switch is_home")
    SORTABLE_FIELDS = strlist("wwpn")

    @command(
        description="List FC soft targets",
        types=(FCSoftTarget.List, FCSwitch, Node, YesNo),
        substitute=dict(
            targets="wwpn",
            ),
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS),
        )
    @render(
        fields=QUERY_FIELDS,
        alignment="left",
        )
    def query(self, _api, targets=None, switch=None, node=None, home=None):
        switch_names_by_id = LazyIdNameMapping(FCSwitch, _api)
        return self._query(_api, targets, switch_id=switch, node_id=node, is_home=home, _hook=partial(self._query_hook, switch_names_by_id=switch_names_by_id))

    def _query_hook(self, targets, switch_names_by_id):
        for target in targets:
            #CLI-1309 product request
            target.wwpn = target.wwpn.lower()
            target.port = FCPort.format_name(target.node_id, target.port_number)
            target.switch = switch_names_by_id[target.switch_id]
        return targets

    @command(
        description="Create soft targets for a given switch",
        types=(FCSwitch, FCSoftTargetCreationQuantity)
        )
    @roles.INFINIDAT.TECHNICIAN
    def create(self, _api, switch, quantity=FCSoftTargetCreationQuantity.DEFAULT_VALUE):
        targets = self._post(_api, None, "create_multiple", data=dict(
            switch_id=switch.id,
            quantity=quantity,
            ))
        targets.sort(key=lambda t: (t.node_id, t.port_number))
        return '{} soft targets created for switch "{}"'.format(len(targets), switch.name)

    @command(
        description="Wipe all soft targets in the system",
        )
    @roles.INFINIDAT
    def wipe(self, _api):
        self._post(_api, None, "wipe")
        return "All FC soft targets in the system successfully wiped"

    @command(
        description="Delete a soft target",
        types=(FCSoftTarget.List,),
        substitute=dict(
            targets="wwpn",
            ),
        )
    @roles.INFINIDAT.TECHNICIAN
    def delete(self, _api, targets):
        return self._delete(_api, targets)
