from munch import Munch
from functools import partial

from infi.shell import commandgroup, render
from infi.shell.types import IPAddress, IPAddressCIDR, YesNo
from infi.shell.utils import strlist
from infi.shell.exceptions import BadValue
from ..types import (
    NONE_CONSTANT,
    NetworkInterfaceName,
    Name,
    EthPort,
    UsableEthPort,
    NetworkInterface,
    NetworkInterfaceType,
    NetworkSpace,
    NetworkSpaceService,
    NetworkRateLimitOrNone,
    Node,
    IPAddressOrNone,
    MTU,
    )
from ..options import QueryOptions
from ..utils import get_patch_panel_ports
from .. import roles, EntityCommands, LazyIdNameMapping, LazyEntityMapping, Null
from . import ConfigCommandsBase

from ...pre_v2_patches import v2_command as command

@commandgroup(parent=ConfigCommandsBase, prefix="ethernet")
class EthernetConfigCommands(object):
    pass

@commandgroup(parent=EthernetConfigCommands, prefix="port")
class EthernetPortCommands(EntityCommands):
    ENTITY_TYPE = EthPort
    QUERY_FIELDS =           strlist("name role rendered_state link_state connection_speed max_speed patch_panel_label")
    DETAILED_QUERY_FIELDS =  QUERY_FIELDS + strlist("hw_addr _name")
    INFINIDAT_QUERY_FIELDS = QUERY_FIELDS + strlist("firmware hw_addr")
    INFINIDAT_COLUMNS_FIELDS = INFINIDAT_QUERY_FIELDS + strlist("_name vendor model")
    QUERY_SUBSTITUTES = dict(
        _name="device_name",
        rendered_state="state",
        link_state="link",
        connection_speed="current_speed",
        patch_panel_label="patch_panel",
        )

    @command(
        description="List system ethernet ports",
        types=(EthPort, Node),
        substitute=dict(
            port="name",
            ),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, infinidat_fields=INFINIDAT_COLUMNS_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render(
        fields=QUERY_FIELDS,
        detailed_fields=DETAILED_QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        infinidat=dict(
            fields=INFINIDAT_QUERY_FIELDS,
            detailed_fields=INFINIDAT_COLUMNS_FIELDS,
            ),
        technician=dict(
            fields=INFINIDAT_QUERY_FIELDS,
            detailed_fields=INFINIDAT_COLUMNS_FIELDS,
            ),
        )
    def query(self, _api, port=None, node=None):
        patch_panel_ports = get_patch_panel_ports("ETH")
        return self._query(_api, port, node_index=node, _hook=partial(self._query_hook, patch_panel_ports=patch_panel_ports))

    def _query_hook(self, ports, patch_panel_ports=None):
        for port in ports:
            port.rendered_state = port.state if not port.state_description \
                                  else "{} ({})".format(port.state, port.state_description)
            port.patch_panel_label = None
            pp_port = None if not patch_panel_ports else patch_panel_ports.get((port.node_index, port.port_number))
            if pp_port:
                port.patch_panel_label = (pp_port.label or pp_port.port_num)
        return ports

    # @command(
    #     description="Enable ethernet port",
    #     types=(EthPort.List,),
    #     substitute=dict(
    #         ports="name",
    #         ),
    #     )
    # def enable(self, _api, ports):
    #     aggregator = _api.create_request_aggregator(
    #         total_success_msg=self._result_message(ports[0], "enabled") if len(ports) == 1 else
    #                           ("Ports enabled: " + ", ".join(p.name for p in ports))
    #         )
    #     with aggregator:
    #         for port in ports:
    #             aggregator.next_task(port.name, "Enabled")
    #             _api.post(EthPort.get_component_uri(port, "enable"))
    #     return aggregator.run()

    # @command(
    #     description="Disable ethernet port",
    #     types=(EthPort.List,),
    #     substitute=dict(
    #         ports="name",
    #         ),
    #     )
    # def disable(self, _api, ports):
    #     aggregator = _api.create_request_aggregator(
    #         total_success_msg=self._result_message(ports[0], "disabled") if len(ports) == 1 else
    #                           ("Ports disabled: " + ", ".join(p.name for p in ports))
    #         )
    #     with aggregator:
    #         for port in ports:
    #             aggregator.next_task(port.name, "Disabled")
    #             _api.post(EthPort.get_component_uri(port, "disable"))
    #     return aggregator.run()

@commandgroup(parent=EthernetConfigCommands, prefix="interface")
class NetworkInterfaceCommands(EntityCommands):
    ENTITY_TYPE = NetworkInterface
    QUERY_FIELDS = strlist("name type node_id port_names network_spaces state")
    SORTABLE_FIELDS = strlist("name node_id")
    QUERY_SUBSTITUTES = dict(
        node_id="node",
        port_names="ports",
        )

    @command(
        description="List existing network interfaces",
        types=(NetworkInterface.List, Node),
        substitute=dict(
            interface="name",
            ),
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES),
        )
    @render(
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        alignment=dict(
            network_spaces="left",
            ),
        )
    def query(self, _api, interface=None, node=None):
        hook = partial(self._query_hook,
            eth_names_to_port_names={(p.node_index, p._name) : p.name for p in EthPort.iter_all(_api)},
            ns_ids_to_names=LazyIdNameMapping(NetworkSpace, _api),
            )
        return self._query(_api, interface, node_id=node, _hook=hook)

    def _query_hook(self, interfaces, eth_names_to_port_names=None, ns_ids_to_names=None):
        for interface in interfaces:
            interface.port_names = [eth_names_to_port_names.get((interface.node_id, p.name), "<N/A>") for p in interface.ports]
            interface.network_spaces = [ns_ids_to_names[ns_id] for ns_id in interface.network_space_ids]
        return interfaces

    @command(
        description="Create network interface",
        types=(NetworkInterfaceName, NetworkInterfaceType, UsableEthPort.List, NetworkRateLimitOrNone, YesNo),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def create(self, _api, name, type, ports, limit=None, repeat_on_all_nodes=False):
        if len(set([p.node_index for p in ports])) > 1:
            raise BadValue, "All ports have to be on the same node."
        port_objects = [dict(name=p._name) for p in ports]
        if repeat_on_all_nodes:
            aggregator = _api.create_request_aggregator()
            with aggregator:
                for node in Node.iter_all(_api):
                    aggregator.next_task("Interface {}".format(name), "created on node-{}".format(node.id))
                    self._create(_api,
                        name="N{}-{}".format(node.id, name),
                        type=type,
                        node_id=node.id,
                        ports=port_objects,
                        rate_limit=limit if limit else Null
                        )
            return aggregator.run()
        else:
            self._create(_api,
                name=name,
                type=type,
                node_id=ports[0].node_index,
                ports=port_objects,
                rate_limit=limit if limit else Null
                )
            return "Interface {} created on node-{}".format(name, ports[0].node_index)

    @command(
        description="Create a single-port network interface (for special purposes only)",
        types=(UsableEthPort,),
        )
    @roles.INFINIDAT
    def create_single_port(self, _api, port):
        return self._create(_api,
            name=port._name,
            type="PORT",
            node_id=port.node_index,
            )

    @command(
        description="Delete network interface",
        types=(NetworkInterface.List,),
        substitute=dict(
            interfaces="name",
            ),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def delete(self, _api, interfaces):
        return self._delete(_api, interfaces)

    @command(
        description="Rename network interface",
        types=(NetworkInterface, Name),
        substitute=dict(
            interface="name",
            ),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def rename(self, _api, interface, new_name):
        return self._rename(_api, interface, new_name)

    @command(
        description="Enable network interface",
        types=(NetworkInterface,),
        substitute=dict(
            interface="name",
            ),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def enable(self, _api, interface):
        self._post(_api, interface, "enable")
        return self._result_message(interface, "enabled")

    @command(
        description="Disable network interface",
        types=(NetworkInterface,),
        substitute=dict(
            interface="name",
            ),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def disable(self, _api, interface):
        self._post(_api, interface, "disable")
        return self._result_message(interface, "disabled")

    @command(
        description="Add ethernet port to network interface",
        types=(NetworkInterface, UsableEthPort.List),
        substitute=dict(
            interface="name",
            ports="port",
            ),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def add_port(self, _api, interface, ports):
        aggregator = _api.create_request_aggregator(
            total_success_msg="Port{} added to {}: {}".format("" if len(ports) == 1 else "s", self._result_subject_and_entity(interface),
                                                              ", ".join(p.name for p in ports)),
            )
        with aggregator:
            for port in ports:
                aggregator.next_task(port.name, "Added")
                self._post(_api, interface, "ports", data=dict(name=port._name))
        return aggregator.run()

    @command(
        description="Remove ethernet port from network interface",
        types=(NetworkInterface, UsableEthPort.List),
        substitute=dict(
            interface="name",
            ports="port",
            ),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN
    def remove_port(self, _api, interface, ports):
        for port in ports:
            if port.node_index != interface.node_id:
                raise BadValue("{} does not contain {}".format(
                    self._result_subject_and_entity(interface, capitalize=True),
                    self._result_subject_and_entity(port, entity_type=UsableEthPort),
                    ))
        aggregator = _api.create_request_aggregator(
            total_success_msg="Port{} removed from {}: {}".format("" if len(ports) == 1 else "s", self._result_subject_and_entity(interface),
                                                                  ", ".join(p.name for p in ports)),
            )
        with aggregator:
            for port in ports:
                aggregator.next_task(port.name, "Removed")
                self._del(_api, interface, "ports", port._name)
        return aggregator.run()

@commandgroup(parent=ConfigCommandsBase, prefix="net_space")
class NetworkSpaceCommands(EntityCommands):
    ENTITY_TYPE = NetworkSpace
    QUERY_FIELDS = strlist("name service network mtu rate_limit gateway ips interface_names automatic_ip_fallback")
    SORTABLE_FIELDS = strlist("name")
    QUERY_SUBSTITUTES = dict(
        services="service",
        automatic_ip_fallback="auto_ip_fallback",
        interface_names="interfaces",
        )
    MGMT_SERVICE_TO_USER_SERVICE = {v : k for k, v in NetworkSpaceService()._get_mapping().iteritems()}

    @command(
        description="List existing network spaces",
        types=(NetworkSpace,),
        substitute=dict(
            ns="name",
            ),
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render(
        fields=QUERY_FIELDS,
        detailed_fields=QUERY_FIELDS,
        detailed_individually=True,
        substitute=QUERY_SUBSTITUTES,
        )
    def query(self, _api, ns=None):
        lazy_interface_name_by_id = LazyIdNameMapping(NetworkInterface, _api)
        return self._query(_api, ns, _hook=partial(self._query_hook, lazy_interface_name_by_id=lazy_interface_name_by_id))

    def _query_hook(self, spaces, lazy_interface_name_by_id):
        for ns in spaces:
            ns.ips = len(ns.ips)
            ns.network, ns.gateway = "{}/{}".format(ns.network_config.network, ns.network_config.netmask), ns.network_config.default_gateway
            ns.interface_names = [lazy_interface_name_by_id[id] for id in ns.interfaces]
            ns.service = self.MGMT_SERVICE_TO_USER_SERVICE.get(ns.service, ns.service)
        return spaces

    @command(
        description="""Sets a bandwidth limitation on a Network Space per node. This limit is configured
        on the Interface used by the Network space to allow traffic shaping and avoid bandwidth starvation""",
        types=(NetworkSpace, NetworkRateLimitOrNone),
        substitute=dict(
            ns="name",
            ),
        )
    @roles.INFINIDAT.ADMIN.TECHNICIAN.POOLADMIN
    def set_rate_limit(self, _api, ns, limit_per_node):
        if limit_per_node is NONE_CONSTANT:
            limit_per_node = Null
        self._update(_api, ns, rate_limit=limit_per_node)
        return self._result_message(ns, "bandwidth limitation (per node): {}".format("Limited to {}Mbps".format(limit_per_node) if limit_per_node is not Null else "Limit removed"))


    @command(
        description="Create network space",
        types=(Name, NetworkSpaceService, NetworkInterface.List, IPAddressCIDR, IPAddress, MTU),
        )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, name, service, interfaces, network, default_gateway=None, mtu=None):
        return self._create(_api,
            name=name,
            service=service,
            network_config=dict(
                netmask=int(network.cidr),  # infi.shell bug, cidr is a string - fixed in v1.12.6
                network=network.ip,
                default_gateway=default_gateway.ip if default_gateway else None,
                ),
            interfaces=[i.id for i in interfaces],
            mtu=mtu,
            )

    @command(
        description="Rename network space",
        types=(NetworkSpace, Name),
        substitute=dict(
            ns="name",
            )
        )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, ns, new_name):
        return self._rename(_api, ns, new_name)

    @command(
        description="Set default gateway",
        types=(NetworkSpace, IPAddressOrNone),
        substitute=dict(
            ns="name",
            )
        )
    @roles.INFINIDAT.ADMIN
    def set_default_gateway(self, _api, ns, default_gateway):
        return self._update(_api, ns,
            network_config=dict(
                netmask=ns.network_config.netmask,
                network=ns.network_config.network,
                default_gateway=None if default_gateway is NONE_CONSTANT else default_gateway.ip,
                ),
            returns=self._result_message(ns, 'default gateway unset' if default_gateway is NONE_CONSTANT
                                             else 'default gateway set to "{}"'.format(default_gateway.ip)),
            )

    @command(
        description="Set network space service",
        types=(NetworkSpace, NetworkSpaceService),
        substitute=dict(
            ns="name",
            )
        )
    @roles.INFINIDAT.ADMIN
    def set_service(self, _api, ns, service):
        return self._update(_api, ns,
            service=service,
            returns=self._result_message(ns, 'service set to "{}"'.format(service)),
            )

    @command(
        description="Set network space port groups",
        types=(NetworkSpace, NetworkInterface.List),
        substitute=dict(
            ns="name",
            )
        )
    @roles.INFINIDAT.ADMIN
    def set_interfaces(self, _api, ns, interfaces):
        return self._update(_api, ns,
            interfaces=[i.id for i in interfaces],
            returns=self._result_message(ns, "interfaces set to " +
                ", ".join('"{}"'.format(i.name) for i in interfaces))
            )

    @command(
        description="Set network space network configuration",
        types=(NetworkSpace, IPAddressCIDR),
        substitute=dict(
            ns="name",
            )
        )
    @roles.INFINIDAT.ADMIN
    def set_network(self, _api, ns, network):
        return self._update(_api, ns,
            network_config=dict(
                netmask=int(network.cidr),  # infi.shell bug, cidr is a string - fixed in v1.12.6
                network=network.ip,
                default_gateway=ns.network_config.default_gateway,
                ),
            returns=self._result_message(ns, 'network set to "{}"'.format(network.ip_and_cidr)),
            )

    @command(
        description="Set network space MTU",
        types=(NetworkSpace, MTU),
        substitute=dict(
            ns="name",
            )
        )
    @roles.INFINIDAT.ADMIN
    def set_mtu(self, _api, ns, mtu):
        return self._update(_api, ns,
            mtu=mtu,
            returns=self._result_message(ns, 'MTU set to {}'.format(mtu)),
            )

    @command(
        description="Delete network space",
        types=(NetworkSpace,),
        substitute=dict(
            ns="name",
            )
        )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, ns):
        return self._delete(_api, ns)

@commandgroup(parent=NetworkSpaceCommands, prefix="ip")
class NetworkSpaceIpCommands(EntityCommands):
    ENTITY_TYPE = NetworkSpace

    QUERY_FIELDS = strlist("network_space ip_address enabled node network_interface reserved")
    QUERY_SUBSTITUTES = dict(
        reserved="replication_management_ip",
        )

    @command(
        description="Query network space IPs",
        types=(NetworkSpace,),
        substitute=dict(
            ns="net_space",
            )
        )
    @render(
        fields=QUERY_FIELDS,
        substitute=QUERY_SUBSTITUTES,
        alignment=dict(
            network_space="left",  # Because "space" is usually used for capacities
            ),
        )
    def query(self, _api, ns=None):
        lazy_interface_by_id = LazyEntityMapping(NetworkInterface, _api, fields=strlist("name node_id"))
        return self._query(_api, ns, _hook=partial(self._query_hook, lazy_interface_by_id=lazy_interface_by_id))

    def _query_hook(self, spaces, lazy_interface_by_id=None):
        for ns in spaces:
            for ip in ns.ips:
                interface = None if not ip.interface_id else lazy_interface_by_id[ip.interface_id]
                yield Munch(
                    network_space=ns.name,
                    ip_address=ip.ip_address,
                    enabled=ip.enabled,
                    reserved=ip.reserved,
                    node=None if not interface else interface.node_id,
                    network_interface=None if not interface else interface.name,
                    )

    def _aggregated_result_message(self, action, ns, ips):
        if len(ips) == 1:
            return 'IP address {} {} in {} {}'.format(ips[0].ip, action, self._result_subject(), self._result_entity_name(ns))
        return "IP addresses {} in {} {}: {}".format(action, self._result_subject(), self._result_entity_name(ns), ", ".join(ip.ip for ip in ips))

    @command(
        description="Create IP address in network space",
        types=(NetworkSpace, IPAddress.List),
        substitute=dict(
            ns="net_space",
            ips="ip_address",
            )
        )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, ns, ips):
        aggregator = _api.create_request_aggregator(
            total_success_msg=self._aggregated_result_message("created", ns, ips)
            )
        with aggregator:
            for ip in ips:
                aggregator.next_task(ip.ip, "Created")
                self._post(_api, ns, "ips", data=ip.ip)
        return aggregator.run()

    def _ensure_ip_exists(self, ns, ip):
        if ip.ip not in (ns_ip.ip_address for ns_ip in ns.ips):
            raise BadValue("{} {} has no such IP address: {}".format(self._result_subject(), self._result_entity_name(ns), ip.ip))

    @command(
        description="Delete IP address from network space",
        types=(NetworkSpace, IPAddress.List),
        substitute=dict(
            ns="net_space",
            ips="ip_address",
            )
        )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, ns, ips):
        for ip in ips:
            self._ensure_ip_exists(ns, ip)
        aggregator = _api.create_request_aggregator(
            total_success_msg=self._aggregated_result_message("deleted", ns, ips)
            )
        with aggregator:
            for ip in ips:
                aggregator.next_task(ip.ip, "Deleted")
                self._del(_api, ns, "ips", ip.ip)
        return aggregator.run()

    @command(
        description="Enable network space IP address",
        types=(NetworkSpace, IPAddress.List),
        substitute=dict(
            ns="net_space",
            ips="ip_address",
            )
        )
    @roles.INFINIDAT.ADMIN
    def enable(self, _api, ns, ips):
        for ip in ips:
            self._ensure_ip_exists(ns, ip)
        aggregator = _api.create_request_aggregator(
            total_success_msg=self._aggregated_result_message("enabled", ns, ips)
            )
        with aggregator:
            for ip in ips:
                aggregator.next_task(ip.ip, "Enabled")
                self._post(_api, ns, "ips", ip.ip, "enable")
        return aggregator.run()

    @command(
        description="Disable network space IP address",
        types=(NetworkSpace, IPAddress.List),
        substitute=dict(
            ns="net_space",
            ips="ip_address",
            )
        )
    @roles.INFINIDAT.ADMIN
    def disable(self, _api, ns, ips):
        for ip in ips:
            self._ensure_ip_exists(ns, ip)
        aggregator = _api.create_request_aggregator(
            total_success_msg=self._aggregated_result_message("disabled", ns, ips)
            )
        with aggregator:
            for ip in ips:
                aggregator.next_task(ip.ip, "Disabled")
                self._post(_api, ns, "ips", ip.ip, "disable")
        return aggregator.run()
