from munch import Munch
import time
from collections import OrderedDict
from functools import partial

from infi.shell import (
    runtime,
    command,
    commandgroup,
    render,
    config,
    get_active_shell,
    BadValue,
    BadArguments,
    )
from infi.shell.types import Timestamp, Str, YesNo, Int, TimeDelta, EmailAddress
from infi.shell.utils import strlist, DictProxy

from ..renderer import CollectionRenderer, CSVRenderer
from ..exceptions import InvalidSystemState, ConnectionFailed, ManagementException
from .types import (
    NONE_CONSTANT,
    Name,
    Event,
    EventInfo,
    EventCode,
    EventCodeOrNone,
    EventLevel,
    EventLevelOrNone,
    EventReporter,
    EventVisibility,
    EventRule,
    NotificationTarget,
    SmtpTarget,
    SnmpTarget,
    SyslogTarget,
    Username,
    )
from .options import QueryOptions, DetailedQuery
from .utils import strip_newline_chars
from . import (
    EntityCommands,
    render_collection,
    roles,
    set_renderer_query_fields,
    LazyEntityMapping,
    )

QUERY_FIELDS                     = strlist("timestamp                           level          code          description")
DETAILED_QUERY_FIELDS            = strlist("timestamp    seq_num                level reporter code username description")
INFINIDAT_DETAILED_QUERY_FIELDS  = strlist("timestamp id seq_num source_node_id level reporter code username description")
DUMP_FIELDS                      = INFINIDAT_DETAILED_QUERY_FIELDS + strlist("affected_entity_id system_version data")

QUERY_SORTABLE_FIELDS = strlist("timestamp seq_num level reporter code username")
INFINIDAT_QUERY_SORTABLE_FIELDS = QUERY_SORTABLE_FIELDS + strlist("id source_node_id")

QUERY_SUBSTITUTES = dict(
    source_node_id="node",
    seq_num="seq",
    )

RENDERER_KWARGS = dict(
    entity=Event,
    fields=QUERY_FIELDS,
    detailed_fields=DETAILED_QUERY_FIELDS,
    substitute=QUERY_SUBSTITUTES,
    alignment=dict(
        id="right",
        source_node_id="right",
        seq_num="right",
        ),
    infinidat=dict(
        fields=QUERY_FIELDS,
        detailed_fields=INFINIDAT_DETAILED_QUERY_FIELDS,
        ),
    )

EVENT_DETAILS_EXCLUDED  = strlist("visibility description_template affected_entity_id")
EVENT_DETAILS_ORDER = INFINIDAT_DETAILED_QUERY_FIELDS + strlist("system_version")

@commandgroup(prefix="event")
class EventCommands(EntityCommands):
    ENTITY_TYPE = Event

    @property
    def _page_size(self):
        return 200 if config.paging and get_active_shell().is_interactive_terminal() else config.api_max_page_size

    @command(
        description="List system events",
        types=(Timestamp, Timestamp, Int, Int, EventCode.List, EventCode.List, EventLevel.List, EventReporter.List, Username.List),
        substitute=dict(from_time="from", to_time="to"),
        options=QueryOptions(fields=DETAILED_QUERY_FIELDS, sortable=QUERY_SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(**RENDERER_KWARGS)
    @roles.exclude.INFINIDAT
    def query(self, _api, from_time=None, to_time=None, from_seq=None, to_seq=None,
                          code=None, exclude=None, level=None, reporter=None, username=None):
        pluck_fields = columns = runtime.options.get("columns")
        if not columns:
            pluck_fields = RENDERER_KWARGS["fields"]
        if ('*' in columns if columns else None) or runtime.options.detailed:
            pluck_fields = RENDERER_KWARGS["detailed_fields"]
        return self._event_query(_api,
            from_time=from_time,
            to_time=to_time,
            from_seq=from_seq,
            to_seq=to_seq,
            include_codes=code,
            exclude_codes=exclude,
            levels=level,
            reporters=reporter,
            usernames=username,
            page_size=self._page_size,
            pluck_fields=pluck_fields,
            hook=self._event_query_hook,
            )

    @command(
        name="query",
        description="List system events",
        types=(Timestamp, Timestamp, Int, Int, Int, Int, EventCode.List, EventCode.List, EventLevel.List, EventReporter.List, Username.List),
        substitute=dict(from_time="from", to_time="to"),
        options=QueryOptions(fields=INFINIDAT_DETAILED_QUERY_FIELDS, sortable=INFINIDAT_QUERY_SORTABLE_FIELDS, substitute=QUERY_SUBSTITUTES, detailed=True),
        )
    @render_collection(**RENDERER_KWARGS)
    @roles.INFINIDAT
    def infinidat__query(self, _api, from_time=None, to_time=None, from_id=None, to_id=None, from_seq=None, to_seq=None,
                                     code=None, exclude=None, level=None, reporter=None, username=None):
        pluck_fields = columns = runtime.options.get("columns")
        if not columns:
            pluck_fields = RENDERER_KWARGS["infinidat"]["fields"]
        if ('*' in columns if columns else None) or runtime.options.detailed:
            pluck_fields = RENDERER_KWARGS["infinidat"]["detailed_fields"]
        return self._event_query(_api,
            from_time=from_time,
            to_time=to_time,
            from_id=from_id,
            to_id=to_id,
            from_seq=from_seq,
            to_seq=to_seq,
            include_codes=code,
            exclude_codes=exclude,
            levels=level,
            reporters=reporter,
            usernames=username,
            page_size=self._page_size,
            pluck_fields=pluck_fields,
            hook=self._event_query_hook,
            )

    @command(
        description="List system events (extended details, CSV format)",
        types=(Timestamp, Timestamp, Int, Int, Int, Int, EventCode.List, EventCode.List, EventLevel.List, EventReporter.List, Username.List, YesNo),
        substitute=dict(from_time="from", to_time="to"),
        )
    @render(
        renderer=CSVRenderer(),
        fields=DUMP_FIELDS,
        )
    @roles.INFINIDAT
    def dump(self, _api, from_time=None, to_time=None, from_id=None, to_id=None, from_seq=None, to_seq=None,
                         code=None, exclude=None, level=None, reporter=None, username=None, allow_newlines=False):
        return self._event_query(_api,
            from_time=from_time,
            to_time=to_time,
            from_id=from_id,
            to_id=to_id,
            from_seq=from_seq,
            to_seq=to_seq,
            include_codes=code,
            exclude_codes=exclude,
            levels=level,
            reporters=reporter,
            usernames=username,
            page_size=self._page_size,
            hook=partial(self._event_dump_hook, allow_newlines=allow_newlines),
            )

    def _event_query(self, _api, from_time=None, to_time=None, from_id=None, to_id=None, from_seq=None, to_seq=None,
                                 include_codes=None, exclude_codes=None, levels=None, reporters=None, usernames=None,
                                 page_size=None, pluck_fields=None, hook=None):
        set_renderer_query_fields()
        filter = self._create_query_filter(_api, from_time, to_time, from_id, to_id, from_seq, to_seq, include_codes, exclude_codes, levels, reporters, usernames)
        sort = self._create_sort(_api, runtime.options.sort) if runtime.options.get("sort") else _api.create_sort().desc("timestamp")
        pluck = None if not pluck_fields else _api.create_pluck(*pluck_fields)
        return self._paged_query(_api, filter=filter, sort=sort, hook=hook, pluck=pluck, page_size=page_size)

    def _create_query_filter(self, _api, from_time=None, to_time=None, from_id=None, to_id=None, from_seq=None, to_seq=None,
                                         include_codes=None, exclude_codes=None, levels=None, reporters=None, usernames=None):
        if include_codes and exclude_codes:
            raise BadArguments("Please specify either code, or exclude, or neither")
        if None not in (from_time, to_time) and from_time > to_time:
            raise BadValue("'from' timestamp must not be greater than 'to' timestamp")
        filter = _api.create_filter()
        filter = self._filter_range(filter, "timestamp", from_time, to_time)
        filter = self._filter_range(filter, "id", from_id, to_id)
        filter = self._filter_range(filter, "seq_num", from_seq, to_seq)
        filter = self._filter_inclusion(filter, "code", include_codes)
        filter = self._filter_exclusion(filter, "code", exclude_codes)
        filter = self._filter_inclusion(filter, "level", levels)
        filter = self._filter_inclusion(filter, "reporter", reporters)
        filter = self._filter_inclusion(filter, "username", usernames)
        return filter

    def _filter_range(self, filter, field, from_value, to_value):
        if from_value is not None:
            if to_value is not None:
                return filter.between(field, from_value, to_value)
            return filter.ge(field, from_value)
        if to_value is not None:
            return filter.le(field, to_value)
        return filter

    def _filter_inclusion(self, filter, field, values):
        if values is not None:
            return filter.member_of(field, *values)
        return filter

    def _filter_exclusion(self, filter, field, values):
        if values is not None:
            return filter.not_member_of(field, *values)
        return filter

    def _event_query_hook(self, events):
        for e in events:
            if e.get("description"):
                e.description = strip_newline_chars(e.description)
        return events

    def _event_dump_hook(self, events, allow_newlines):
        for e in events:
            if not allow_newlines and e.get("description"):
                e.description = strip_newline_chars(e.description)
            if e.get("data"):
                data = sorted((field.name, field.value) for field in e.data
                              # Show only non-strings, and non-empty strings (omit empty strings)
                              if not isinstance(field.value, basestring) or field.value
                              # Hide SysConf data (in e.g heartbeat events)
                              and field.type.lower() != "sysconf")
                e.data = " ".join("{}={}".format(key, value if allow_newlines else strip_newline_chars(value))
                                  for key, value in data)
        return events

    @command(
        description="Display an individual event",
        types=(Int,),
        )
    @render_collection(
        individually=True,
        excluded=EVENT_DETAILS_EXCLUDED,
        order=EVENT_DETAILS_ORDER,
        substitute=QUERY_SUBSTITUTES,
        )
    @roles.exclude.INFINIDAT
    def details(self, _api, seq):
        return self._details(_api, seq=seq)

    @command(
        name="details",
        description="Display an individual event",
        types=(Int, Int),
        )
    @render_collection(
        individually=True,
        excluded=EVENT_DETAILS_EXCLUDED,
        order=EVENT_DETAILS_ORDER,
        substitute=QUERY_SUBSTITUTES,
        )
    @roles.INFINIDAT
    def infinidat__details(self, _api, id=None, seq=None):
        if (id, seq).count(None) != 1:
            raise BadArguments('Either "id" or "seq" must be specified')
        return self._details(_api, id=id, seq=seq)

    def _details(self, _api, id=None, seq=None):
        assert (id, seq).count(None) == 1, 'Either id or seq must be specified'
        event = Event.get_by_id(_api, id) if seq is None else Event.get_by_seq(_api, seq)
        data = event.data
        event = OrderedDict(sorted((key, value) for key, value in event.iteritems() if key not in ("visibility", "description_template", "data")))
        for param in sorted(data, key=lambda p: p.name):
            event['data: ' + param.name] = param.value
        return [event]

    def _format_constants(self, _api, clazz, column_name):
        return [Munch({column_name:value}) for value in clazz.iter_all(_api)]

    CODES_QUERY_FIELDS = strlist("code level reporter description")
    CODES_INFINIDAT_QUERY_FIELDS = strlist("code level reporter description visibility")
    @command(
        description="List all event codes",
        types=(EventCode.List, EventLevel.List, EventReporter.List),
        options=[QueryOptions(fields=CODES_QUERY_FIELDS, infinidat_fields=CODES_INFINIDAT_QUERY_FIELDS, sortable=CODES_INFINIDAT_QUERY_FIELDS, detailed=True)],
        )
    @render_collection(
        entity=Event,
        fields=CODES_QUERY_FIELDS,
        detailed_fields=CODES_QUERY_FIELDS,
        detailed_individually=True,
        infinidat=dict(
            fields=CODES_INFINIDAT_QUERY_FIELDS,
            detailed_fields=CODES_INFINIDAT_QUERY_FIELDS,
            ),
        )
    def codes(self, _api, code=None, level=None, reporter=None):
        hook = None
        attrs = dict(code=code, level=level, reporter=reporter)
        attrs = {attr : value for attr, value in attrs.iteritems() if value}
        if any(attrs.itervalues()):
            hook = partial(self._codes_query_hook, attrs=attrs)
        return self._query(_api, _entity_type=EventInfo, _hook=hook)

    def _codes_query_hook(self, events, attrs):
        return [e for e in events if all(e[attr] in values for attr, values in attrs.iteritems())]

    @command(
        description="List all event levels",
        )
    @render(
        header=False,
        )
    def levels(self, _api):
        return self._format_constants(_api, EventLevel, 'level')

    @command(
        description="List all event reporters",
        )
    @render(
        header=False,
        )
    def reporters(self, _api):
        return self._format_constants(_api, EventReporter, 'reporter')

    @command
    @render(
        header=False,
        )
    @roles.INFINIDAT
    def visibilities(self, _api):
        return self._format_constants(_api, EventVisibility, 'visibility')

    @command(
        description="Create a new custom event",
        types=(EventLevel, Str),
        )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, level, description):
        _api.post("events/custom", dict(
            visibility="CUSTOMER",
            level=level,
            description_template=description,
            ))
        return "Event created."

    @command(
        description="Create a heartbeat event",
        )
    @roles.INFINIDAT
    def heartbeat(self, _api):
        _api.internal.post("events/heartbeat")
        return "Heartbeat event created."

    @command(
        description="Monitor system events in real time",
        types=(TimeDelta, EventCode.List, EventCode.List, EventLevel.List, EventReporter.List, Username.List),
        options=(DetailedQuery(),),
        )
    def watch(self, _api, interval=2, code=None, exclude=None, level=None, reporter=None, username=None):
        self._error_interval = 10
        self._time_passed = 0
        renderer_kwargs = RENDERER_KWARGS
        if runtime.options.detailed:
            renderer_kwargs = DictProxy(RENDERER_KWARGS, dict(
                fields=DETAILED_QUERY_FIELDS,
                infinidat=DictProxy(RENDERER_KWARGS["infinidat"], dict(
                    fields=INFINIDAT_DETAILED_QUERY_FIELDS,
                    ))
                ))
        render = partial(
            CollectionRenderer().render,
            stream=get_active_shell().output_stream,
            header=False,
            **renderer_kwargs
            )
        filter = self._create_query_filter(_api, include_codes=code, exclude_codes=exclude, levels=level, reporters=reporter, usernames=username)
        events = list(reversed(Event.paged_query(_api,
            sort=_api.create_sort().desc("timestamp"),
            filter=filter,
            page_size=10,
            hook=self._event_query_hook,
            ).get_page(1).objects))
        last_timestamp = 0
        while True:
            try:
                if events:
                    last_timestamp = events[-1].timestamp
                    render(events, header=False)
                time.sleep(interval)
                events = Event.get_all(_api,
                    sort=_api.create_sort().asc("timestamp"),
                    filter=filter.gt("timestamp", last_timestamp),
                    hook=self._event_query_hook,
                    )
                self._time_passed = 0
            except (ConnectionFailed, ManagementException) as e:
                time.sleep(self._error_interval)
                self._time_passed += self._error_interval
                if self._time_passed > 500:
                    raise e


@commandgroup(prefix="event.rule")
class EventRuleCommands(EntityCommands):
    ENTITY_TYPE = EventRule
    VISIBILITIES = Munch(INFINIDAT="INFINIDAT", CUSTOMER="CUSTOMER")


    EVENT_RULE_QUERY_FIELDS = strlist("name target server event_level recipients")
    DETAILED_EVENT_RULE_QUERY_FIELDS = strlist("name target server event_level recipients event_visibility exclude_events include_events event_code")

    @command(
        description="List existing event rules",
        types=(EventRule,),
        substitute=dict(rule="name"),
        options=QueryOptions(fields=EVENT_RULE_QUERY_FIELDS, infinidat_fields=DETAILED_EVENT_RULE_QUERY_FIELDS, detailed=True),
        )
    @render(
        fields=EVENT_RULE_QUERY_FIELDS,
        detailed_individually=True,
        detailed_fields=DETAILED_EVENT_RULE_QUERY_FIELDS,
        substitute=dict(
            event_level="level",
            include_events="include",
            exclude_events="exclude",
            ),
        )
    def query(self, _api, rule=None):
        targets_by_id = LazyEntityMapping(NotificationTarget, _api, fields=("name", "protocol"))
        return self._query(_api, rule, _hook=partial(self._query_hook, targets_by_id=targets_by_id))

    def _query_hook(self, rules, targets_by_id):
        for r in rules:
            target = targets_by_id[r.target_id]
            r.target = target.protocol
            r.server = target.name
            r.recipients = None
            if r.target_parameters and r.target_parameters.get("recipients"):
                r.recipients = r.target_parameters.recipients
            yield r

    @command(
        description="Create a new event notification rule",
        types=(Name, EventLevel.List, EventCode.List, EventCode.List, EmailAddress.List, SnmpTarget, SyslogTarget),
        )
    @roles.ADMIN
    def create(self, _api, name, level=None, include=None, exclude=None, email=None, snmp=None, syslog=None):
        if (email or None, snmp or None, syslog or None).count(None) != 2:
            raise BadArguments("Please specify either email, snmp, or syslog.")
        smtp = None if not email else self._get_customer_smtp_target(_api)
        return self._create_rule(_api, name, level=level, include=include, exclude=exclude, email=email, smtp=smtp, snmp=snmp, syslog=syslog)

    @command(
        name="create",
        description="Create a new event notification rule",
        types=(Name, EventLevel.List, EventCode.List, EventCode.List, EmailAddress.List, SmtpTarget, SnmpTarget, SyslogTarget),
        )
    @roles.INFINIDAT
    def infinidat__create(self, _api, name, level=None, include=None, exclude=None, email=None, smtp=None, snmp=None, syslog=None):
        if any((email, smtp, snmp, syslog)) and (syslog and snmp or ((syslog or snmp) and (email or smtp)) or (email, smtp).count(None) == 1):
            raise BadArguments("Please specify either snmp, syslog, or both email and smtp.")
        return self._create_rule(_api, name, level=level, include=include, exclude=exclude, email=email, smtp=smtp, snmp=snmp, syslog=syslog)

    def _create_rule(self, _api, name, level=None, include=None, exclude=None, email=None, smtp=None, snmp=None, syslog=None):
        if not (level or include):
            raise BadArguments("Please specify either level, include or both")
        params = {}
        for p in (
            self._get_rule_criteria_params(level, include, exclude),
            self._get_rule_target_params(email, smtp, snmp, syslog),
            ):
            params.update(p)
        return self._create(_api, name=name, **params)

    @command(
        description="Modify existing event notification rule",
        types=(EventRule, EventLevelOrNone.List, EventCodeOrNone.List, EventCodeOrNone.List, EmailAddress.List, SnmpTarget, SyslogTarget),
        substitute=dict(rule="name"),
        )
    @roles.ADMIN
    def modify(self, _api, rule, level=None, include=None, exclude=None, email=None, snmp=None, syslog=None):
        if any((email, snmp, syslog)) and (email or None, snmp or None, syslog or None).count(None) != 2:
            raise BadArguments("Please specify either email, snmp, or syslog.")
        smtp = None if not email else self._get_customer_smtp_target(_api)
        return self._modify(_api, rule, level=level, include=include, exclude=exclude, email=email, smtp=smtp, snmp=snmp, syslog=syslog)

    @command(
        name="modify",
        description="Modify existing event notification rule",
        types=(EventRule, EventLevelOrNone.List, EventCodeOrNone.List, EventCodeOrNone.List, EmailAddress.List, SmtpTarget, SnmpTarget, SyslogTarget),
        substitute=dict(rule="name"),
        )
    @roles.INFINIDAT
    def infinidat__modify(self, _api, rule, level=None, include=None, exclude=None, email=None, smtp=None, snmp=None, syslog=None):
        if any((email, smtp, snmp, syslog)) and (syslog and snmp or ((syslog or snmp) and (email or smtp)) or (email, smtp).count(None) == 1):
            raise BadArguments("Please specify either snmp, syslog, or both email and smtp.")
        return self._modify(_api, rule, level=level, include=include, exclude=exclude, email=email, smtp=smtp, snmp=snmp, syslog=syslog)

    def _get_customer_smtp_target(self, _api):
        smtp_targets = [t for t in SmtpTarget.get_all(_api)
                        if t.get("visibility", "").upper() != "INFINIDAT"]
        if len(smtp_targets) < 1:
            raise InvalidSystemState("SMTP notification target not configured - please contact customer support")
        if len(smtp_targets) > 1:
            raise InvalidSystemState("Multiple SMTP targets exist - please contact customer support")
        return smtp_targets[0]

    def _modify(self, _api, rule, level=None, include=None, exclude=None, email=None, smtp=None, snmp=None, syslog=None):
        assert (email or None, smtp or None).count(None) != 1, "email and smtp must be specified together"
        if self._is_explicit_none_value(level) and not rule.include_events and (not include or self._is_explicit_none_value(include)):
            raise BadArguments("Cannot reset level for a rule that does not specify included events")
        if self._is_explicit_none_value(include) and not rule.event_level and (not level or self._is_explicit_none_value(level)):
            raise BadArguments("Cannot reset included events for a rule that does not specify any level")
        params = {}
        if any((level, include, exclude)):
            for arg, argtype in (
                ("level", EventLevelOrNone),
                ("include", EventCodeOrNone),
                ("exclude", EventCodeOrNone),
                ):
                items = locals()[arg]
                if items and len(items) > 1 and any(item is NONE_CONSTANT for item in items):
                    raise BadValue('When specifying {0} in "{1}", {0} must be the sole value'.format(argtype().get_constant_name(NONE_CONSTANT), arg))
            params.update(self._get_rule_criteria_params(level, include, exclude))
        if any((email, smtp, snmp, syslog)):
            params.update(self._get_rule_target_params(email, smtp, snmp, syslog))
        if not params:
            raise BadArguments("At least one more argument must be specified")
        field_update_order = ["event_level"] if not params.get("include_events") else ["include_events"]
        return self._update(_api, rule, _field_update_order=field_update_order, **params)

    def _is_explicit_none_value(self, list_arg):
        return list_arg is not None and len(list_arg) == 1 and list_arg[0] is NONE_CONSTANT

    def _get_rule_criteria_params(self, level, include, exclude):
        get_list_value = lambda lst: [] if self._is_explicit_none_value(lst) else lst
        return OrderedDict(
            event_level=get_list_value(level),
            include_events=get_list_value(include),
            exclude_events=get_list_value(exclude),
            )

    def _get_rule_target_params(self, email, smtp, snmp, syslog):
        params = dict(
            target_parameters=dict(),
            )
        for target in (smtp, snmp, syslog):
            if target:
                params["target_id"] = target.id
                # Set rule's visibility according to target's visibility. Note
                # that the rule's visibility implies whether or not internal
                # events are also processed by the rule.
                params["event_visibility"] = ["CUSTOMER"]
                if target.visibility == "INFINIDAT":
                    params["event_visibility"].append("INFINIDAT")
                break
        if email:
            params["target_parameters"] = dict(
                recipients=email,
                )
        return params

    @command(
        description="Rename existing event notification rule",
        types=(EventRule, Name),
        substitute=dict(rule="name"),
        )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, rule, new_name):
        return self._rename(_api, rule, new_name)

    @command(
        description="Delete event notification rule",
        types=(EventRule,),
        substitute=dict(rule="name"),
        )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, rule):
        return self._delete(_api, rule)
