from munch import Munch

from infi.shell import commandgroup, command, render
from infi.shell.utils import strlist

from ..system import system
from .types import Node, NodeDrive
from .options import QueryOptions
from .utils import strip_newline_chars
from . import EntityCommands, roles

QUERY_FIELDS = strlist("name state core_state core_role mgmt_state mgmt_role")
INFINIDAT_QUERY_FIELDS = QUERY_FIELDS + strlist("vendor model")

DRIVE_QUERY_FIELDS = strlist("name type state state_description")
INFINIDAT_DRIVE_QUERY_FIELDS = strlist("name type state vendor model serial_number firmware state_description")

@commandgroup(prefix="node")
class NodeCommands(EntityCommands):
    ENTITY_TYPE = Node

    @command(
        description="List system nodes",
        options=QueryOptions(fields=QUERY_FIELDS, infinidat_fields=INFINIDAT_QUERY_FIELDS),
        )
    @render(
        fields=QUERY_FIELDS,
        infinidat=dict(
            fields=INFINIDAT_QUERY_FIELDS,
            ),
        technician=dict(
            fields=INFINIDAT_QUERY_FIELDS,
            ),
        )
    def query(self, _api):
        return self._query(_api, _hook=self._query_hook)

    def _query_hook(self, nodes):
        for node in nodes:
            services = Munch((s.name, s) for s in node.services)
            yield Munch(name=node.name,
                        state=node.state,
                        vendor=node.vendor,
                        model=node.model,
                        core_state=services.core.state,
                        core_role=services.core.role,
                        mgmt_state=services.mgmt.state,
                        mgmt_role=services.mgmt.role,
                        )

    @command(
        description="Initiate node activation process",
        types=(Node,),
        )
    @roles.INFINIDAT.TECHNICIAN
    def activate(self, _api, node):
        _api.post("components/nodes/{}/phase_in".format(node.id))
        return "Activating {}".format(node.name)

    @command(
        description="Initiate node deactivation process",
        types=(Node,),
        )
    @roles.INFINIDAT.TECHNICIAN
    def deactivate(self, _api, node):
        _api.post("components/nodes/{}/phase_out".format(node.id))
        return "Deactivating {}".format(node.name)

@commandgroup(prefix="drive", parent=NodeCommands)
class NodeDriveCommands(EntityCommands):
    ENTITY_TYPE = NodeDrive

    @command(
        description="List node drives",
        types=(NodeDrive, Node),
        substitute=dict(
            drive="name",
            ),
        options=QueryOptions(fields=DRIVE_QUERY_FIELDS, infinidat_fields=INFINIDAT_DRIVE_QUERY_FIELDS),
        )
    @render(
        fields=DRIVE_QUERY_FIELDS,
        infinidat=dict(
            fields=INFINIDAT_DRIVE_QUERY_FIELDS
            ),
        )
    def query(self, _api, drive=None, node=None):
        return self._query(_api, drive, node_index=node, _hook=self._node_drive_query_hook)

    def _node_drive_query_hook(self, drives):
        for drive in drives:
            if drive.get("state_description"):
                drive.state_description = strip_newline_chars(drive.state_description)
            if drive.state == "UNCONFIGURED" and system.user.roles.is_customer():
                continue
            yield drive
