from functools import partial
from collections import defaultdict

from infi.shell import commandgroup, command, render, get_active_shell
from infi.shell.types import EmailAddress, Str
from infi.shell.utils import strlist
from infi.shell.exceptions import CommandFailed, CommandAborted, BadArguments

from ...utils import input_password
from ..system import system
from .types import (
    Name,
    LocalUser,
    LdapGroup,
    PoolAdminUser,
    PoolAdminGroup,
    UsableUserRole,
    MappableUserRole,
    Pool,
    UsersRepository,
    LdapDN,
    )
from .options import QueryOptions
from .utils import user_has_role, is_privileged_user, increased_api_timeout
from . import EntityCommands, roles, LazyIdNameMapping

class UserCommandsBase(object):

    def _get_pool_names_by_user_id(self, _api):
        pool_names_by_user_id = defaultdict(list)
        for p in Pool.iter_all(_api):
            for uid in p.owners:
                names = pool_names_by_user_id[uid]
                names.append(p.name)
                names.sort()
        return pool_names_by_user_id

    def _get_user_pools(self, user, pool_names_by_user_id):
        return "<all pools>" if is_privileged_user(user) else pool_names_by_user_id[user.id]

    def _get_subject(self, user):
        return '{} "{}"'.format((LocalUser if user.type.lower() == "local" else LdapGroup).get_entity_name(), user.name)

    def _add_pool(self, _api, user, pool):
        _api.post(Pool.get_uri(pool.id, "owners", user.id))
        return 'Granted provisioning privileges on pool "{}" to {}' \
               .format(pool.name, self._get_subject(user))

    def _remove_pool(self, _api, user, pool):
        _api.delete(Pool.get_uri(pool.id, "owners", user.id))
        return 'Revoked provisioning privileges on pool "{}" from {}' \
               .format(pool.name, self._get_subject(user))

@commandgroup(prefix="user")
class UserCommands(UserCommandsBase, EntityCommands):
    ENTITY_TYPE = LocalUser
    QUERY_FIELDS = strlist("name role pools")
    SORTABLE_FIELDS = strlist("name role")

    @command(
        description="List existing local users",
        types=(LocalUser.List,),
        substitute=dict(user="name"),
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS),
        )
    @render(
        fields=QUERY_FIELDS,
        )
    def query(self, _api, user=None):
        hook = partial(self._query_hook, pool_names_by_user_id=self._get_pool_names_by_user_id(_api))
        return self._query(_api, user, _hook=hook)

    def _query_hook(self, users, pool_names_by_user_id):
        for u in users:
            u.pools = self._get_user_pools(u, pool_names_by_user_id)
        return users

    def _input_new_password(self, username=None):
        if not (system.user.roles.is_infinidat() or system.user.roles.is_admin()):
            # Check current password
            if username is not None and not system.check_credentials(username, input_password("Current password: ")):
                if username == system.user.username:
                    get_active_shell().error("This is not your current password, terminating shell")
                    raise EOFError
                raise CommandFailed("This is not the current password, please try again")
        password = input_password(prompt="New password: ")
        if not password.strip():
            raise CommandAborted("Password is empty, aborting")
        if password != input_password(prompt="Re-enter new password: "):
            raise CommandFailed("Passwords differ from one another, please try again")
        return password

    @command(
        description="Create a new local user",
        types=(Name, EmailAddress, UsableUserRole, Str),
        )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, name, email, role, password=None):
        if password is None:
            password = input_password('Please enter a password for user "{}": '.format(name),
                                      interaction_exc_type=BadArguments, interaction_error="Please specify a password")
        if not password:
            raise CommandAborted("Password is empty")
        return self._create(_api,
            type=LocalUser.USER_TYPE,
            name=name,
            password=password,
            email=email,
            role=role,
            )

    @command(
        description="Change local user username",
        types=(LocalUser, Str,),
        substitute=dict(user="name"),
        )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, user, new_name):
        return self._rename(_api, user, new_name)

    @command(
        description="Change local user role",
        types=(LocalUser, UsableUserRole,),
        substitute=dict(user="name"),
        )
    @roles.INFINIDAT.ADMIN
    def change_role(self, _api, user, role):
        if user_has_role(user, role):
            return self._result_message(user, "role is already {}".format(role))
        self._update(_api, user, role=role)
        return self._result_message(user, "role changed from {} to {}".format(user.roles[0], role))

    @command(
        description="Change local user email",
        types=(LocalUser, EmailAddress,),
        substitute=dict(user="name"),
        )
    @roles.INFINIDAT.ADMIN
    def change_email(self, _api, user, email):
        self._update(_api, user, email=email)
        return self._result_message(user, "email changed from {} to {}".format(user.email, email))

    @command(
        description="Delete local user",
        types=(LocalUser,),
        substitute=dict(user="name"),
        )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, user):
        return self._delete(_api, user)

    @command(
        description="Interactively reset password of local user",
        substitute=dict(user="name"),
        )
    @roles.exclude.INFINIDAT
    def change_password(self, _api):
        return self._change_password(_api)

    @command(
        name="change_password",
        description="Interactively reset password of local user",
        types=(LocalUser,),
        substitute=dict(user="name"),
        )
    @roles.INFINIDAT
    def infinidat__change_password(self, _api, user=None):
        return self._change_password(_api, user)

    def _change_password(self, _api, user=None):
        user = user or LocalUser.get_by_name(_api, system.user.username)
        password = self._input_new_password(user.name)
        self._update(_api, user, password=password)
        if system.user.username == user.name:
            system.login(user.name, password)
        return self._result_message(user, "password changed")

    @command(
        description="Grant pool provisioning privileges to a pool administrator",
        types=(PoolAdminUser, Pool,),
        substitute=dict(user="name"),
        )
    @roles.INFINIDAT.ADMIN
    def add_pool(self, _api, user, pool):
        return self._add_pool(_api, user, pool)

    @command(
        description="Revoke pool provisioning privileges from a pool administrator",
        types=(PoolAdminUser, Pool,),
        substitute=dict(user="name"),
        )
    @roles.INFINIDAT.ADMIN
    def remove_pool(self, _api, user, pool):
        return self._remove_pool(_api, user, pool)

@commandgroup(parent=UserCommands, prefix="group")
class LdapRoleMappingCommands(UserCommandsBase, EntityCommands):
    ENTITY_TYPE = LdapGroup
    QUERY_FIELDS = strlist("name role ldap dn pools")
    SORTABLE_FIELDS = strlist("name role dn")

    @command(
        description="List existing LDAP groups",
        types=(LdapGroup,),
        substitute=dict(
            group="name",
            ),
        options=QueryOptions(fields=QUERY_FIELDS, sortable=SORTABLE_FIELDS),
        )
    @render(
        fields=QUERY_FIELDS,
        )
    def query(self, _api, group=None):
        hook = partial(self._query_hook,
            ldap_name_by_id=LazyIdNameMapping(UsersRepository, _api),
            pool_names_by_user_id=self._get_pool_names_by_user_id(_api),
            )
        return self._query(_api, group, _hook=hook)

    def _query_hook(self, groups, ldap_name_by_id, pool_names_by_user_id):
        for g in groups:
            g.ldap = ldap_name_by_id[g.ldap_id]
            g.pools = self._get_user_pools(g, pool_names_by_user_id)
        return groups

    @command(
        description="Map LDAP or Active Directory group to user role",
        types=(Name, MappableUserRole, UsersRepository, LdapDN),
        )
    @roles.INFINIDAT.ADMIN
    def create(self, _api, name, role, ldap, dn):
        with increased_api_timeout():
            return self._create(_api,
                type=LdapGroup.USER_TYPE,
                name=name,
                roles=[role],
                ldap_id=ldap,
                dn=dn,
                )

    @command(
        description="Change LDAP or Active Directory group name",
        types=(LdapGroup, Str,),
        substitute=dict(group="name"),
        )
    @roles.INFINIDAT.ADMIN
    def rename(self, _api, group, new_name):
        return self._rename(_api, group, new_name)

    @command(
        description="Unmap LDAP or Active Directory group from user role",
        types=(LdapGroup,),
        substitute=dict(
            group="name",
            ),
        )
    @roles.INFINIDAT.ADMIN
    def delete(self, _api, group):
        return self._delete(_api, group)

    @command(
        description="Change LDAP or Active Directory group role",
        types=(LdapGroup, MappableUserRole,),
        substitute=dict(group="name"),
        )
    @roles.INFINIDAT.ADMIN
    def change_role(self, _api, group, role):
        if user_has_role(group, role):
            return self._result_message(group, "role is already {}".format(role))
        self._update(_api, group, role=role)
        return self._result_message(group, "role changed from {} to {}".format(group.roles[0], role))

    @command(
        description="Grant pool provisioning privileges to an LDAP group of pool administrators",
        types=(PoolAdminGroup, Pool,),
        substitute=dict(user="name"),
        )
    @roles.INFINIDAT.ADMIN
    def add_pool(self, _api, user, pool):
        return self._add_pool(_api, user, pool)

    @command(
        description="Revoke pool provisioning privileges from an LDAP group of pool administrators",
        types=(PoolAdminGroup, Pool,),
        substitute=dict(user="name"),
        )
    @roles.INFINIDAT.ADMIN
    def remove_pool(self, _api, user, pool):
        return self._remove_pool(_api, user, pool)
