from __future__ import print_function

# Import patches first
from . import patches  # noqa

# Import everything else
import sys
import traceback
import colorama
from functools import partial

from infi.shell import (
    Commands,
    Parser, ShellParser,
    config,
    )
from infi.shell.exceptions import FilesystemException
from infi.shell.formatters import format_exception

from .. import debug
from ..main import BASE_URI  # Imported for testing  # noqa
from ..main import (
    REST_BASE_URI,
    INTERNAL_BASE_URI,
    output,
    error,
    get_api_base_url,
    get_history_filename,
    new_client_id,
    get_login_password,
    )
from ..constants import PRODUCT_NAME
from ..environment import variables as env
from ..plugins import plugins
from ..utils import post_mortem
from ..shell import InfiniboxShell, InfiniboxExecutor
from ..exceptions import InvalidOptions

from . import raw_mode, __commands_manifest__
from .api import InfiniboxAPI
from .args_processor import InfiniboxArgsProcessor
from .formatters import InfinidatExceptionFormatter
from .constants import ROOT_USER
# Due to cyclic dependencies, types has to be imported before renderers
from .commands import types as ___  # noqa
from .commands.options import DebugOption, AutoApprove
from .renderer import InfiniboxRenderer, toggle_csv_mode, toggle_json_mode
from .system.system import System, system as system_context
from .system.user import Roles

from .exceptions import (
    ConnectionFailed,
    ServiceTemporarilyUnavailable,
    InternalServerError,
    ManagementException,
    UnexpectedServerResponse,
    APIFailure,
    )

# Import command modules so that they register in the manifest
from .manifest import manifest

from .pre_v2_patches import pre_v2_patches

COMMANDS = [cls() for cls in manifest.iter_manifest(__commands_manifest__)]
assert COMMANDS, "No commands registered in manifest"

GLOBAL_OPTIONS = (
    DebugOption(),
    AutoApprove(),
    )

def get_default_headers():
    headers = dict()
    for header, var in (
            ("X-Forwarded-For", "X_FORWARDED_FOR"),
            ):
        if env.is_set(var):
            headers[header] = env.get(var)
    return headers

def _prompt_escape(colors):
    return "\x01" + colors + "\x02"

_prompt_template = "{user}@{system}"
_colored_prompt_template = "{user_color}{{user}}{at_color}@{system_color}{{system}}{end_color}".format(
    user_color=_prompt_escape(colorama.Fore.BLUE + colorama.Style.BRIGHT),
    at_color=_prompt_escape(colorama.Fore.RESET + colorama.Style.BRIGHT),
    system_color=_prompt_escape(colorama.Fore.GREEN + colorama.Style.NORMAL),
    end_color=_prompt_escape(colorama.Style.RESET_ALL),
    )

def prompt_hook(shell, address, system):
    template = _colored_prompt_template if shell.has_colors() else _prompt_template
    return template.format(user=system.user.username, system=address)

_prompt_sep = "> "
_colored_prompt_sep = "{sep_color}>{end_color} ".format(
    sep_color=_prompt_escape(colorama.Fore.RESET + colorama.Style.BRIGHT),
    end_color=_prompt_escape(colorama.Style.RESET_ALL),
    )

def prompt_sep_hook(shell):
    return _colored_prompt_sep if shell.has_colors() else _prompt_sep

def run(options):
    if options.raw:
        raw_mode.toggle()
    try:
        try:
            plugins.apply_config()

            is_interactive = options.script is None
            base_api = InfiniboxAPI(
                    get_api_base_url(options.address, options.port),
                    new_client_id(),
                    headers=get_default_headers(),
                    encoding=options.encoding,
                    )
            api = base_api.get_sub_resource(REST_BASE_URI, internal=base_api.get_sub_resource(INTERNAL_BASE_URI))

            system = System(api)
            system_context.__push_context__(system)

            if is_interactive:
                output("Connecting to {}...", options.address)

            # The no-auth mode comes to work around having to authenticate when
            # using the shell from within the system's OS shell (SSH)
            if env.is_flagged("NO_AUTH"):
                system._set_user(ROOT_USER, [Roles.INFINIDAT])
            elif env.is_set("EMULATE_LOGIN"):
                login_info = env.get("EMULATE_LOGIN").split(",", 2)
                if len(login_info) == 3:
                    [username, password, roles] = login_info
                elif len(login_info) == 1:
                    username, password, [roles] = config.username, get_login_password(), login_info
                else:
                    raise InvalidOptions('Environment variable {}_EMULATE_LOGIN format is invalid: "{}".\n'
                                         'Expected format (1): "username,password,role[,role,...]"\n'
                                         '                (2): "role"'
                                         .format(env.get_prefix(), login_info))
                system._set_user(username, roles.split(","))
                api.set_basic_auth(username, password)
            else:
                system.login(config.username, get_login_password())  # Also sets API basic auth

            pre_v2_patches(api)

            commands = Commands(
                    *(list(COMMANDS) + list(plugins.create_commands())),
                    global_options=GLOBAL_OPTIONS,
                    args_processor=InfiniboxArgsProcessor(),
                    default_renderer=InfiniboxRenderer(),
                    api=api
                )

            parser_kwargs = dict(commands=commands, encoding=options.encoding)
            _color_str = config.color.lower()
            _color = True if _color_str == "always" else False if _color_str == "never" else None
            out_stream = sys.stdout
            common_kwargs = dict(
                stream=out_stream,
                paging=config.paging,
                exc_formatter=InfinidatExceptionFormatter(),
                color=_color,
                errors_color=_color,
                )

            def prepare(shell):
                if config.log_file:
                    shell.start_logging(config.log_file)
                if config.csv_output:
                    toggle_csv_mode(shell)
                if config.json_output:
                    toggle_json_mode(shell)

            result_code = 0

            if is_interactive:
                shell = InfiniboxShell(
                    parser=ShellParser(**parser_kwargs),
                    history_filename=get_history_filename(options.address),
                    encoding=options.encoding,
                    prompt=partial(prompt_hook, address=options.address, system=system),
                    prompt_sep=prompt_sep_hook,
                    welcome_msg=options.welcome_msg,
                    exit_msg=options.exit_msg,
                    error_stream=out_stream,  # In interactive mode stderr is not used
                    **common_kwargs
                    )
            else:
                shell = InfiniboxExecutor(
                    parser=Parser(**parser_kwargs),
                    keep_going=config.keep_going,
                    quiet=config.quiet,
                    **common_kwargs
                    )

            prepare(shell)
            plugins.before_any_session(shell)

            if is_interactive:
                plugins.before_interactive_session(shell)
                shell.start()
                plugins.after_interactive_session(shell)
            else:
                plugins.before_script_session(shell)
                success = shell.execute(options.script)
                result_code = 0 if success else 2
                plugins.after_script_session(shell, success)

            plugins.after_any_session(shell)

            return result_code
        except:
            # In debug mode, stack trace should be printed for all exceptions
            if env.is_flagged("DEBUG"):
                traceback.print_exc()
            raise
    except (ConnectionFailed, ManagementException, UnexpectedServerResponse) as e:
        error("Cannot connect to {} system{}: {}", PRODUCT_NAME, " {}".format(system._system_name) if system._system_name else "",  str(e))
        return 2
    except ServiceTemporarilyUnavailable:
        error("{} temporarily unavailable", PRODUCT_NAME)
        return 2
    except InternalServerError:
        error("An unknown error occurred on the server side: {}", format_exception())
        return 2
    except APIFailure as e:
        msg = "Failed to communicate with server due to product or version mismatch"
        if env.is_flagged("DEBUG") or debug.is_on():
            error(msg + ": {}", format_exception())
        else:
            error(msg)
        return 2
    except FilesystemException as e:
        error(str(e))
        return 1
    except (InvalidOptions, KeyboardInterrupt):
        raise
    except:
        # Save exc_info now, in case other exceptions arise before post mortem
        exc_info = sys.exc_info()
        # For unknown exceptions, stack trace should be printed
        if not env.is_flagged("DEBUG"):
            if debug.is_on():
                traceback.print_exc()
            error("An unexpected error has occurred - please contact customer support")
        # In debug mode, perform post mortem
        else:
            post_mortem(exc_info)
        return 1
